# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Reoptimize THC factors using a combination of BFGS and adaagrad including a
penalty parameter to regularize the lambda value as described in
https://arxiv.org/abs/2302.05531

The entrypoint for a user should be the function kpoint_thc_via_isdf which can
be used as

>>> krhf_inst = scf.KRHF(cell, kpts)
>>> krmp_inst = scf.KMP2(krhf_inst)
>>> Lchol = pyscf_chol_from_df(krmp_inst)
>>> nthc = cthc * nmo
>>> thc_factors = kpoint_thc_via_isdf(krhf_inst, Lchol, nthc)
"""

# pylint: disable=wrong-import-position
import math
import time
from typing import Tuple, Union

import h5py
import numpy as np
import numpy.typing as npt
from pyscf.pbc import scf
from scipy.optimize import minimize

from jax.config import config
config.update("jax_enable_x64", True)

import jax
import jax.numpy as jnp
import jax.typing as jnpt

from openfermion.resource_estimates.thc.utils import adagrad
from openfermion.resource_estimates.pbc.thc.factorizations.isdf import (
    KPointTHC,
    solve_kmeans_kpisdf,
)
from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)


def load_thc_factors(chkfile_name: str) -> KPointTHC:
    """Load THC factors from a checkpoint file

    Args:
        chkfile_name: Filename containing THC factors.

    Returns:
        kthc: KPointISDF object built from chkfile_name.
    """
    xi = None
    with h5py.File(chkfile_name, "r") as fh5:
        chi = fh5["chi"][:]
        g_mapping = fh5["g_mapping"][:]
        num_kpts = g_mapping.shape[0]
        zeta = np.zeros((num_kpts,), dtype=object)
        if "xi" in list(fh5.keys()):
            xi = fh5["xi"][:]
        else:
            xi = None
        for iq in range(g_mapping.shape[0]):
            zeta[iq] = fh5[f"zeta_{iq}"][:]
    return KPointTHC(xi=xi, zeta=zeta, g_mapping=g_mapping, chi=chi)


def save_thc_factors(
        chkfile_name: str,
        chi: npt.NDArray,
        zeta: npt.NDArray,
        gpq_map: npt.NDArray,
        xi: Union[npt.NDArray, None] = None,
) -> None:
    """Write THC factors to file

    Args:
        chkfile_name: Filename to write to.
        chi: THC leaf tensor.
        zeta: THC central tensor.
        gpq_map: maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        xi: Interpolating vectors (optional, Default None).
    """
    num_kpts = chi.shape[0]
    with h5py.File(chkfile_name, "w") as fh5:
        fh5["chi"] = chi
        fh5["g_mapping"] = gpq_map
        if xi is not None:
            fh5["xi"] = xi
        for iq in range(num_kpts):
            fh5[f"zeta_{iq}"] = zeta[iq]


def get_zeta_size(zeta: npt.NDArray) -> int:
    """zeta (THC central tensor) is not contiguous so this helper function
    returns its size

    Args:
        zeta: THC central tensor

    Returns:
        zeta_size: Number of elements in zeta
    """
    return sum([z.size for z in zeta])


def unpack_thc_factors(
        xcur: npt.NDArray,
        num_thc: int,
        num_orb: int,
        num_kpts: int,
        num_g_per_q: list,
) -> Tuple[npt.NDArray, npt.NDArray]:
    """Unpack THC factors from flattened array used for reoptimization.

    Args:
        xcur: Flattened array containing k-point THC factors.
        num_thc: THC rank.
        num_orb: Number of orbitals.
        num_kpts: Number of kpoints.
        num_g_per_q: number of g vectors per q vector.

    Returns:
        chi: THC leaf tensor.
        zeta: THC central tensor.
    """
    # leaf tensor (num_kpts, num_orb, num_thc)
    chi_size = num_kpts * num_orb * num_thc
    chi_real = xcur[:chi_size].reshape(num_kpts, num_orb, num_thc)
    chi_imag = xcur[chi_size:2 * chi_size].reshape(num_kpts, num_orb, num_thc)
    chi = chi_real + 1j * chi_imag
    zeta_packed = xcur[2 * chi_size:]
    zeta = []
    start = 0
    for iq in range(num_kpts):
        num_g = num_g_per_q[iq]
        size = num_g * num_g * num_thc * num_thc
        zeta_real = zeta_packed[start:start + size].reshape(
            (num_g, num_g, num_thc, num_thc))
        zeta_imag = zeta_packed[start + size:start + 2 * size].reshape(
            (num_g, num_g, num_thc, num_thc))
        zeta.append(zeta_real + 1j * zeta_imag)
        start += 2 * size
    return chi, zeta


def pack_thc_factors(chi: npt.NDArray, zeta: npt.NDArray,
                     buffer: npt.NDArray) -> None:
    """Pack THC factors into flattened array used for reoptimization.

    Args:
        chi: THC leaf tensor.
        zeta: THC central tensor.
        buffer: Flattened array containing k-point THC factors. Modified inplace
    """
    assert len(chi.shape) == 3
    buffer[:chi.size] = chi.real.ravel()
    buffer[chi.size:2 * chi.size] = chi.imag.ravel()
    start = 2 * chi.size
    num_kpts = len(zeta)
    for iq in range(num_kpts):
        size = zeta[iq].size
        buffer[start:start + size] = zeta[iq].real.ravel()
        buffer[start + size:start + 2 * size] = zeta[iq].imag.ravel()
        start += 2 * size


@jax.jit
def compute_objective_batched(
        chis: Tuple[jnpt.ArrayLike, jnpt.ArrayLike, jnpt.ArrayLike, jnpt.
                    ArrayLike],
        zetas: jnpt.ArrayLike,
        chols: Tuple[jnpt.ArrayLike, jnpt.ArrayLike],
        norm_factors: Tuple[jnpt.ArrayLike, jnpt.ArrayLike, jnpt.
                            ArrayLike, jnpt.ArrayLike],
        num_kpts: int,
        penalty_param: float = 0.0,
) -> float:
    """Compute THC objective function.

    Batches evaluation over kpts.

    Args:
        chis: THC leaf tensor.
        zetas: THC central tensor.
        chols: Cholesky factors definining 'exact' eris.
        norm_factors: THC normalization factors.
        num_kpts: Number of k-points.
        penalty_param: Penalty parameter.

    Returns:
        objective: THC objective function
    """
    eri_thc = jnp.einsum(
        "Jpm,Jqm,Jmn,Jrn,Jsn->Jpqrs",
        chis[0].conj(),
        chis[1],
        zetas,
        chis[2].conj(),
        chis[3],
        optimize=True,
    )
    eri_ref = jnp.einsum("Jnpq,Jnrs->Jpqrs", chols[0], chols[1], optimize=True)
    deri = (eri_thc - eri_ref) / num_kpts
    norm_left = norm_factors[0] * norm_factors[1]
    norm_right = norm_factors[2] * norm_factors[3]
    mpq_normalized = (jnp.einsum(
        "JP,JPQ,JQ->JPQ", norm_left, zetas, norm_right, optimize=True) /
                      num_kpts)

    lambda_z = jnp.sum(jnp.einsum("jpq->j", 0.5 * jnp.abs(mpq_normalized))**2.0)

    res = 0.5 * jnp.sum((jnp.abs(deri))**2) + penalty_param * lambda_z
    return res


def prepare_batched_data_indx_arrays(
        momentum_map: npt.NDArray,
        gpq_map: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray]:
    r"""Create arrays to batch over.

    Flatten sum_q sum_{k,k_prime} -> sum_q \sum_{indx} and pack momentum
    conserving indices and central tensors so we can sum over indx efficiently.

    Args:
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.

    Returns:
        indx_pqrs: momentum conserving k-point indices.
        zetas: Batches Central tensors.
    """
    num_kpts = momentum_map.shape[0]
    indx_pqrs = np.zeros((num_kpts, num_kpts**2, 4), dtype=jnp.int32)
    zetas = np.zeros((num_kpts, num_kpts**2, 2), dtype=jnp.int32)
    for iq in range(num_kpts):
        indx = 0
        for ik in range(num_kpts):
            ik_minus_q = momentum_map[iq, ik]
            gpq = gpq_map[iq, ik]
            for ik_prime in range(num_kpts):
                ik_prime_minus_q = momentum_map[iq, ik_prime]
                gsr = gpq_map[iq, ik_prime]
                indx_pqrs[iq, indx] = [
                    ik,
                    ik_minus_q,
                    ik_prime_minus_q,
                    ik_prime,
                ]
                zetas[iq, indx] = [gpq, gsr]
                indx += 1
    return indx_pqrs, zetas


@jax.jit
def get_batched_data_1indx(array: jnpt.ArrayLike,
                           indx: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """Helper function to extract entries of array given another array.

    Args:
        array: Array to index
        indx: Indexing array

    Retuns:
        indexed_array: i.e. array[indx]
    """
    return array[indx]


@jax.jit
def get_batched_data_2indx(array: jnpt.ArrayLike, indxa: jnpt.ArrayLike,
                           indxb: jnpt.ArrayLike) -> jnpt.ArrayLike:
    """Helper function to extract entries of 2D array given another array

    Args:
        array: Array to index
        indxa: Indexing array
        indxb: Indexing array

    Retuns:
        indexed_array: i.e. array[indxa, indxb]
    """
    return array[indxa, indxb]


def thc_objective_regularized_batched(
        xcur: jnpt.ArrayLike,
        num_orb: int,
        num_thc: int,
        momentum_map: npt.NDArray,
        gpq_map: npt.NDArray,
        chol: jnpt.ArrayLike,
        indx_arrays: Tuple[jnpt.ArrayLike, jnpt.ArrayLike],
        batch_size: int,
        penalty_param=0.0,
) -> float:
    """Compute THC objective function.

    Here we batch over multiple k-point indices for improved GPU efficiency.

    Args:
        xcur: Flattened array containing k-point THC factors.
        num_orb: Number of orbitals.
        num_thc: THC rank.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.
        indx_arrays: Batched index arrays (see prepare_batched_data_indx_arrays)
        batch_size: Size of each batch of data. Should be in range
        [1, num_kpts**2]. penalty_param: Penalty param if computing regularized
        cost function.

    Returns:
        objective: THC objective function
    """
    num_kpts = momentum_map.shape[0]
    num_g_per_q = [len(np.unique(gq)) for gq in gpq_map]
    chi, zeta = unpack_thc_factors(xcur, num_thc, num_orb, num_kpts,
                                   num_g_per_q)
    # Normalization factor, no factor of sqrt as there are 4 chis in total when
    # building ERI.
    norm_kp = jnp.einsum("kpP,kpP->kP", chi.conj(), chi, optimize=True)**0.5
    num_batches = math.ceil(num_kpts**2 / batch_size)

    indx_pqrs, indx_zeta = indx_arrays
    objective = 0.0
    for iq in range(num_kpts):
        for ibatch in range(num_batches):
            start = ibatch * batch_size
            end = (ibatch + 1) * batch_size
            chi_p = get_batched_data_1indx(chi, indx_pqrs[iq, start:end, 0])
            chi_q = get_batched_data_1indx(chi, indx_pqrs[iq, start:end, 1])
            chi_r = get_batched_data_1indx(chi, indx_pqrs[iq, start:end, 2])
            chi_s = get_batched_data_1indx(chi, indx_pqrs[iq, start:end, 3])
            norm_k1 = get_batched_data_1indx(norm_kp,
                                             indx_pqrs[iq, start:end, 0])
            norm_k2 = get_batched_data_1indx(norm_kp,
                                             indx_pqrs[iq, start:end, 1])
            norm_k3 = get_batched_data_1indx(norm_kp,
                                             indx_pqrs[iq, start:end, 2])
            norm_k4 = get_batched_data_1indx(norm_kp,
                                             indx_pqrs[iq, start:end, 3])
            zeta_batch = get_batched_data_2indx(
                zeta[iq],
                indx_zeta[iq, start:end, 0],
                indx_zeta[iq, start:end, 1],
            )
            chol_batch_pq = get_batched_data_2indx(chol,
                                                   indx_pqrs[iq, start:end, 0],
                                                   indx_pqrs[iq, start:end, 1])
            chol_batch_rs = get_batched_data_2indx(chol,
                                                   indx_pqrs[iq, start:end, 2],
                                                   indx_pqrs[iq, start:end, 3])
            objective += compute_objective_batched(
                (chi_p, chi_q, chi_r, chi_s),
                zeta_batch,
                (chol_batch_pq, chol_batch_rs),
                (norm_k1, norm_k2, norm_k3, norm_k4),
                num_kpts,
                penalty_param=penalty_param,
            )
    return objective


def thc_objective_regularized(
        xcur,
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        chol,
        penalty_param=0.0,
):
    """Compute THC objective function. Non-batched version.

    Args:
        xcur: Flattened array containing k-point THC factors.
        num_orb: Number of orbitals.
        num_thc: THC rank.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.
        penalty_param: Penalty param if computing regularized cost function.

    Returns:
        objective: THC objective function
    """
    res = 0.0
    num_kpts = momentum_map.shape[0]
    num_g_per_q = [len(np.unique(GQ)) for GQ in gpq_map]
    chi, zeta = unpack_thc_factors(xcur, num_thc, num_orb, num_kpts,
                                   num_g_per_q)
    num_kpts = momentum_map.shape[0]
    norm_kP = jnp.einsum("kpP,kpP->kP", chi.conj(), chi, optimize=True)**0.5
    for iq in range(num_kpts):
        for ik in range(num_kpts):
            ik_minus_q = momentum_map[iq, ik]
            Gpq = gpq_map[iq, ik]
            for ik_prime in range(num_kpts):
                ik_prime_minus_q = momentum_map[iq, ik_prime]
                Gsr = gpq_map[iq, ik_prime]
                eri_thc = jnp.einsum(
                    "pm,qm,mn,rn,sn->pqrs",
                    chi[ik].conj(),
                    chi[ik_minus_q],
                    zeta[iq][Gpq, Gsr],
                    chi[ik_prime_minus_q].conj(),
                    chi[ik_prime],
                )
                eri_ref = jnp.einsum(
                    "npq,nsr->pqrs",
                    chol[ik, ik_minus_q],
                    chol[ik_prime, ik_prime_minus_q].conj(),
                )
                deri = (eri_thc - eri_ref) / num_kpts
                norm_left = norm_kP[ik] * norm_kP[ik_minus_q]
                norm_right = norm_kP[ik_prime_minus_q] * norm_kP[ik_prime]
                MPQ_normalized = (jnp.einsum("P,PQ,Q->PQ", norm_left,
                                             zeta[iq][Gpq, Gsr], norm_right) /
                                  num_kpts)

                lambda_z = 0.5 * jnp.sum(jnp.abs(MPQ_normalized))
                res += 0.5 * jnp.sum(
                    (jnp.abs(deri))**2) + penalty_param * (lambda_z**2)

    return res


def lbfgsb_opt_kpthc_l2reg(
        chi: npt.NDArray,
        zeta: npt.NDArray,
        momentum_map: npt.NDArray,
        gpq_map: npt.NDArray,
        chol: npt.NDArray,
        chkfile_name: Union[str, None] = None,
        maxiter: int = 150_000,
        disp_freq: int = 98,
        penalty_param: Union[float, None] = None,
) -> Tuple[npt.NDArray, float]:
    """Least-squares fit of two-electron integral tensors with  L-BFGS-B

    Uses l2-regularization of lambda penalty function.

    Args:
        chi: THC leaf tensor.
        zeta: THC central tensor.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.
        batch_size: Size of each batch of data. Should be in range
            [1, num_kpts**2].
        penalty_param: Penalty param if computing regularized cost function.
        chkfile_name: Filename to store intermediate state of optimization to.
        maxiter: Max L-BFGS-B iteration.
        disp_freq: L-BFGS-B disp_freq.
        penalty_param: Paramter to penalize optimization by one-norm of
            Hamiltonian. If None it is determined automatically through trying
            to balance the two terms in the objective function.

    Returns:
        objective: THC objective function
    """
    initial_guess = np.zeros(2 * (chi.size + get_zeta_size(zeta)),
                             dtype=np.float64)
    pack_thc_factors(chi, zeta, initial_guess)
    assert len(chi.shape) == 3
    assert len(zeta[0].shape) == 4
    num_kpts = chi.shape[0]
    num_orb = chi.shape[1]
    num_thc = chi.shape[-1]
    assert zeta[0].shape[-1] == num_thc
    assert zeta[0].shape[-2] == num_thc
    print(initial_guess)
    loss = thc_objective_regularized(
        jnp.array(initial_guess),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        penalty_param=0.0,
    )
    reg_loss = thc_objective_regularized(
        jnp.array(initial_guess),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        penalty_param=1.0,
    )
    # set penalty
    lambda_z = (reg_loss - loss)**0.5
    if penalty_param is None:
        # loss + lambda_z^2 - loss
        penalty_param = loss / lambda_z
    print("loss {}".format(loss))
    print("lambda_z {}".format(lambda_z))
    print("penalty_param {}".format(penalty_param))

    # L-BFGS-B optimization
    thc_grad = jax.grad(thc_objective_regularized, argnums=[0])
    print("Initial Grad")
    print(
        thc_grad(
            jnp.array(initial_guess),
            num_orb,
            num_thc,
            momentum_map,
            gpq_map,
            jnp.array(chol),
            penalty_param,
        ))
    res = minimize(
        thc_objective_regularized,
        initial_guess,
        args=(
            num_orb,
            num_thc,
            momentum_map,
            gpq_map,
            jnp.array(chol),
            penalty_param,
        ),
        jac=thc_grad,
        method="L-BFGS-B",
        options={
            "disp": disp_freq > 0,
            "iprint": disp_freq,
            "maxiter": maxiter,
        },
    )

    params = np.array(res.x)
    loss = thc_objective_regularized(
        jnp.array(res.x),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        penalty_param=0.0,
    )
    if chkfile_name is not None:
        num_g_per_q = [len(np.unique(GQ)) for GQ in gpq_map]
        chi, zeta = unpack_thc_factors(params, num_thc, num_orb, num_kpts,
                                       num_g_per_q)
        save_thc_factors(chkfile_name, chi, zeta, gpq_map)
    return np.array(params), loss


def lbfgsb_opt_kpthc_l2reg_batched(
        chi: npt.NDArray,
        zeta: npt.NDArray,
        momentum_map: npt.NDArray,
        gpq_map: npt.NDArray,
        chol: npt.NDArray,
        chkfile_name: Union[str, None] = None,
        maxiter: int = 150_000,
        disp_freq: int = 98,
        penalty_param: Union[float, None] = None,
        batch_size: Union[int, None] = None,
) -> Tuple[npt.NDArray, float]:
    """Least-squares fit of two-electron integral tensors with  L-BFGS-B.

    Uses l2-regularization of lambda. This version batches over multiple
    k-points which may be faster on GPUs.

    Args:
        chi: THC leaf tensor.
        zeta: THC central tensor.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.
        batch_size: Size of each batch of data. Should be in range
            [1, num_kpts**2].
        penalty_param: Penalty param if computing regularized cost function.
        chkfile_name: Filename to store intermediate state of optimization to.
        maxiter: Max L-BFGS-B iteration.
        disp_freq: L-BFGS-B disp_freq.
        penalty_param: Paramter to penalize optimization by one-norm of
            Hamiltonian. If None it is determined automatically through trying
            to balance the two terms in the objective function.
        batch_size: Number of k-points-pairs to batch over. Default num_kpts**2.

    Returns:
        objective: THC objective function
    """
    initial_guess = np.zeros(2 * (chi.size + get_zeta_size(zeta)),
                             dtype=np.float64)
    pack_thc_factors(chi, zeta, initial_guess)
    assert len(chi.shape) == 3
    assert len(zeta[0].shape) == 4
    num_kpts = chi.shape[0]
    num_orb = chi.shape[1]
    num_thc = chi.shape[-1]
    assert zeta[0].shape[-1] == num_thc
    assert zeta[0].shape[-2] == num_thc
    if batch_size is None:
        batch_size = num_kpts**2
    indx_arrays = prepare_batched_data_indx_arrays(momentum_map, gpq_map)
    data_amount = batch_size * (
        4 * num_orb * num_thc + num_thc * num_thc  # chi[p,m] + zeta[m,n]
    )
    data_size_gb = (data_amount * 16) / (1024**3)
    print(f"Batch size in GB: {data_size_gb}")
    loss = thc_objective_regularized_batched(
        jnp.array(initial_guess),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        indx_arrays,
        batch_size,
        penalty_param=0.0,
    )
    start = time.time()
    reg_loss = thc_objective_regularized_batched(
        jnp.array(initial_guess),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        indx_arrays,
        batch_size,
        penalty_param=1.0,
    )
    print("Time to evaluate loss function : {:.4f}".format(time.time() - start))
    print("loss {}".format(loss))
    # set penalty
    lambda_z = (reg_loss - loss)**0.5
    if penalty_param is None:
        # loss + lambda_z^2 - loss
        penalty_param = loss / lambda_z
    print("lambda_z {}".format(lambda_z))
    print("penalty_param {}".format(penalty_param))

    # L-BFGS-B optimization
    thc_grad = jax.grad(thc_objective_regularized_batched, argnums=[0])
    print("Initial Grad")
    start = time.time()
    print(
        thc_grad(
            jnp.array(initial_guess),
            num_orb,
            num_thc,
            momentum_map,
            gpq_map,
            jnp.array(chol),
            indx_arrays,
            batch_size,
            penalty_param,
        ))
    print("# Time to evaluate gradient: {:.4f}".format(time.time() - start))
    res = minimize(
        thc_objective_regularized_batched,
        initial_guess,
        args=(
            num_orb,
            num_thc,
            momentum_map,
            gpq_map,
            jnp.array(chol),
            indx_arrays,
            batch_size,
            penalty_param,
        ),
        jac=thc_grad,
        method="L-BFGS-B",
        options={
            "disp": disp_freq > 0,
            "iprint": disp_freq,
            "maxiter": maxiter,
        },
    )
    loss = thc_objective_regularized_batched(
        jnp.array(res.x),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        indx_arrays,
        batch_size,
        penalty_param=0.0,
    )

    params = np.array(res.x)
    if chkfile_name is not None:
        num_g_per_q = [len(np.unique(gq)) for gq in gpq_map]
        chi, zeta = unpack_thc_factors(params, num_thc, num_orb, num_kpts,
                                       num_g_per_q)
        save_thc_factors(chkfile_name, chi, zeta, gpq_map)
    return np.array(params), loss


def adagrad_opt_kpthc_batched(
        chi,
        zeta,
        momentum_map,
        gpq_map,
        chol,
        batch_size=None,
        chkfile_name=None,
        stepsize=0.01,
        momentum=0.9,
        maxiter=50_000,
        gtol=1.0e-5,
) -> Tuple[npt.NDArray, float]:
    """Adagrad optimization of THC factors.

    THC optimization usually starts with BFGS and then is completed with Adagrad
    or other first order solver.

    Args:
        chi: THC leaf tensor.
        zeta: THC central tensor.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.
        batch_size: Size of each batch of data. Should be in range
            [1, num_kpts**2].
        chkfile_name: Filename to store intermediate state of optimization to.
        maxiter: Max L-BFGS-B iteration.
        disp_freq: L-BFGS-B disp_freq.
        penalty_param: Paramter to penalize optimization by one-norm of
            Hamiltonian. If None it is determined automatically through trying
            to balance the two terms in the objective function.
        batch_size: Number of k-points-pairs to batch over. Default num_kpts**2.

    Returns:
        objective: THC objective function
    """
    assert len(chi.shape) == 3
    assert len(zeta[0].shape) == 4
    num_kpts = chi.shape[0]
    num_orb = chi.shape[1]
    num_thc = chi.shape[-1]
    assert zeta[0].shape[-1] == num_thc
    assert zeta[0].shape[-2] == num_thc
    # set initial guess
    initial_guess = np.zeros(2 * (chi.size + get_zeta_size(zeta)),
                             dtype=np.float64)
    pack_thc_factors(chi, zeta, initial_guess)
    opt_init, opt_update, get_params = adagrad(step_size=stepsize,
                                               momentum=momentum)
    opt_state = opt_init(initial_guess)
    thc_grad = jax.grad(thc_objective_regularized_batched, argnums=[0])

    if batch_size is None:
        batch_size = num_kpts**2
    indx_arrays = prepare_batched_data_indx_arrays(momentum_map, gpq_map)

    def update(i, opt_state):
        params = get_params(opt_state)
        gradient = thc_grad(
            params,
            num_orb,
            num_thc,
            momentum_map,
            gpq_map,
            chol,
            indx_arrays,
            batch_size,
        )
        grad_norm_l1 = np.linalg.norm(gradient[0], ord=1)
        return opt_update(i, gradient[0], opt_state), grad_norm_l1

    for t in range(maxiter):
        opt_state, grad_l1 = update(t, opt_state)
        params = get_params(opt_state)
        if t % 500 == 0:
            fval = thc_objective_regularized_batched(
                params,
                num_orb,
                num_thc,
                momentum_map,
                gpq_map,
                chol,
                indx_arrays,
                batch_size,
            )
            outline = "Objective val {: 5.15f}".format(fval)
            outline += "\tGrad L1-norm {: 5.15f}".format(grad_l1)
            print(outline)
        if grad_l1 <= gtol:
            # break out of loop
            # which sends to save
            break
    else:
        print("Maximum number of iterations reached")
    # save results before returning
    x = np.array(params)
    loss = thc_objective_regularized_batched(
        params,
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        chol,
        indx_arrays,
        batch_size,
    )
    if chkfile_name is not None:
        num_g_per_q = [len(np.unique(GQ)) for GQ in gpq_map]
        chi, zeta = unpack_thc_factors(x, num_thc, num_orb, num_kpts,
                                       num_g_per_q)
        save_thc_factors(chkfile_name, chi, zeta, gpq_map)
    return params, loss


def make_contiguous_cholesky(cholesky: npt.NDArray) -> npt.NDArray:
    """It is convenient for optimization to make the cholesky array contiguous.
    This function truncates and auxiliary index that is greater than the minimum
    number of auxiliary vectors.

    Args:
        cholesky: Cholesky vectors

    Returns:
        cholesk_contg: Contiguous array of cholesky vectors.
    """
    num_kpts = len(cholesky)
    num_mo = cholesky[0, 0].shape[-1]
    if cholesky.dtype == object:
        # Jax requires contiguous arrays so just truncate naux if it's not
        # uniform hopefully shouldn't affect results dramatically as from
        # experience the naux amount only varies by a few % per k-point
        # Alternatively todo: padd with zeros
        min_naux = min([cholesky[k1, k1].shape[0] for k1 in range(num_kpts)])
        cholesky_contiguous = np.zeros(
            (
                num_kpts,
                num_kpts,
                min_naux,
                num_mo,
                num_mo,
            ),
            dtype=np.complex128,
        )
        for ik1 in range(num_kpts):
            for ik2 in range(num_kpts):
                cholesky_contiguous[ik1, ik2] = cholesky[ik1, ik2][:min_naux]
    else:
        cholesky_contiguous = cholesky

    return cholesky_contiguous


def compute_isdf_loss(chi, zeta, momentum_map, gpq_map, chol):
    """Helper function to compute ISDF loss.

    Args:
        chi: THC leaf tensor.
        zeta: THC central tensor.
        momentum_map: momentum transfer mapping. map[iQ, ik_p] -> ik_q;
            (kpts[ikp] - kpts[ikq])%G = kpts[iQ].
        gpq_map: Maps momentum conserving tuples of kpoints to reciprocal
            lattice vectors in THC central tensor.
        chol: Cholesky factors definining 'exact' eris.

    Returns:
        loss: ISDF loss in ERIs.
    """
    initial_guess = np.zeros(2 * (chi.size + get_zeta_size(zeta)),
                             dtype=np.float64)
    pack_thc_factors(chi, zeta, initial_guess)
    assert len(chi.shape) == 3
    assert len(zeta[0].shape) == 4
    num_orb = chi.shape[1]
    num_thc = chi.shape[-1]
    loss = thc_objective_regularized(
        jnp.array(initial_guess),
        num_orb,
        num_thc,
        momentum_map,
        gpq_map,
        jnp.array(chol),
        penalty_param=0.0,
    )
    return loss


def kpoint_thc_via_isdf(
        kmf: scf.RHF,
        cholesky: npt.NDArray,
        num_thc: int,
        perform_bfgs_opt: bool = True,
        perform_adagrad_opt: bool = True,
        bfgs_maxiter: int = 3000,
        adagrad_maxiter: int = 3000,
        checkpoint_basename: str = "thc",
        save_checkpoints: bool = False,
        use_batched_algos: bool = True,
        penalty_param: Union[None, float] = None,
        batch_size: Union[None, bool] = None,
        max_kmeans_iteration: int = 500,
        verbose: bool = False,
        initial_guess: Union[None, KPointTHC] = None,
        isdf_density_guess: bool = False,
) -> Tuple[KPointTHC, dict]:
    """
    Solve k-point THC using ISDF as an initial guess.

    Arguments:
        kmf: instance of pyscf.pbc.KRHF object. Must be using FFTDF density
            fitting for integrals.
        cholesky: 3-index cholesky integrals needed for BFGS optimization.
        num_thc: THC dimensions (int), usually nthc = c_thc * n, where n is the
            number of spatial orbitals in the unit cell and c_thc is some
            poisiitve integer (typically <= 15).
        perform_bfgs_opt: Perform subsequent BFGS optimization of THC
            factors?
        perform_adagrad_opt: Perform subsequent adagrad optimization of THC
            factors? This is performed after BFGD if perform_bfgs_opt is True.
        bfgs_maxiter: Maximum iteration for adagrad optimization.
        adagrad_maxiter: Maximum iteration for adagrad optimization.
        save_checkpoints: Whether to save checkpoint files or not (which will
            store THC factors. For each step we store checkpoints as
            {checkpoint_basename}_{thc_method}.h5, where thc_method is one of
            the strings (isdf, bfgs or adagrad).
        checkpoint_basename: Base name for checkpoint files. string,
            default "thc".
        use_batched_algos: Whether to use batched algorithms which may be
            faster but have higher memory cost. Bool. Default True.
        penalty_param: Penalty parameter for l2 regularization. Float.
            Default None.
        max_kmeans_iteration: Maximum number of iterations for KMeansCVT
            step. int. Default 500.
        verbose: Print information? Bool, default False.

    Returns:
        kthc: k-point THC factors
        info: dictionary of losses for each stage of factorization.
    """
    # Perform initial ISDF calculation of THC factors
    info = {}
    start = time.time()
    if initial_guess is not None:
        kpt_thc = initial_guess
    else:
        kpt_thc = solve_kmeans_kpisdf(
            kmf,
            num_thc,
            single_translation=False,
            verbose=verbose,
            max_kmeans_iteration=max_kmeans_iteration,
            use_density_guess=isdf_density_guess,
        )
    if verbose:
        print("Time for generating initial guess {:.4f}".format(time.time() -
                                                                start))
    num_mo = kmf.mo_coeff[0].shape[-1]
    num_kpts = len(kmf.kpts)
    chi, zeta, g_mapping = kpt_thc.chi, kpt_thc.zeta, kpt_thc.g_mapping
    if save_checkpoints:
        chkfile_name = f"{checkpoint_basename}_isdf.h5"
        save_thc_factors(chkfile_name, chi, zeta, g_mapping, kpt_thc.xi)
    momentum_map = build_momentum_transfer_mapping(kmf.cell, kmf.kpts)
    if cholesky is not None:
        cholesky_contiguous = make_contiguous_cholesky(cholesky)
        info["loss_isdf"] = compute_isdf_loss(chi, zeta, momentum_map,
                                              g_mapping, cholesky_contiguous)
    start = time.time()
    if perform_bfgs_opt:
        if save_checkpoints:
            chkfile_name = f"{checkpoint_basename}_bfgs.h5"
        else:
            chkfile_name = None
        if use_batched_algos:
            opt_params, loss_bfgs = lbfgsb_opt_kpthc_l2reg_batched(
                chi,
                zeta,
                momentum_map,
                g_mapping,
                cholesky_contiguous,
                chkfile_name=chkfile_name,
                maxiter=bfgs_maxiter,
                penalty_param=penalty_param,
                batch_size=batch_size,
                disp_freq=(98 if verbose else -1),
            )
            info["loss_bfgs"] = loss_bfgs
        else:
            opt_params, loss_bfgs = lbfgsb_opt_kpthc_l2reg(
                chi,
                zeta,
                momentum_map,
                g_mapping,
                cholesky_contiguous,
                chkfile_name=chkfile_name,
                maxiter=bfgs_maxiter,
                penalty_param=penalty_param,
                disp_freq=(98 if verbose else -1),
            )
            info["loss_bfgs"] = loss_bfgs
        num_g_per_q = [len(np.unique(GQ)) for GQ in g_mapping]
        chi, zeta = unpack_thc_factors(opt_params, num_thc, num_mo, num_kpts,
                                       num_g_per_q)
    if verbose:
        print("Time for BFGS {:.4f}".format(time.time() - start))
    start = time.time()
    if perform_adagrad_opt:
        if save_checkpoints:
            chkfile_name = f"{checkpoint_basename}_adagrad.h5"
        else:
            chkfile_name = None
        if use_batched_algos:
            opt_params, loss_ada = adagrad_opt_kpthc_batched(
                chi,
                zeta,
                momentum_map,
                g_mapping,
                cholesky_contiguous,
                chkfile_name=chkfile_name,
                maxiter=adagrad_maxiter,
                batch_size=batch_size,
            )
            info["loss_adagrad"] = loss_ada
        num_g_per_q = [len(np.unique(GQ)) for GQ in g_mapping]
        chi, zeta = unpack_thc_factors(opt_params, num_thc, num_mo, num_kpts,
                                       num_g_per_q)
    if verbose:
        print("Time for ADAGRAD {:.4f}".format(time.time() - start))
    result = KPointTHC(chi=chi, zeta=zeta, g_mapping=g_mapping, xi=kpt_thc.xi)
    return result, info



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Determine costs for DF decomposition"""
from typing import Tuple, Union, List
import numpy as np
from numpy.lib.scimath import arccos, arcsin
from sympy import factorint

from openfermion.resource_estimates.utils import QI

from openfermion.resource_estimates.pbc.resources import (
    ResourceEstimates,
    QR3,
)


def compute_beta_for_resources(num_spin_orbs: int, num_kpts: int,
                               de_for_qpe: float):
    """Compute beta (number of bits for controlled rotations).

    Uses expression from https://arxiv.org/pdf/2007.14460.pdf.

    Args:
        num_spin_orbs: Number of spin orbitals (per k-point)
        num_kpts: Number of k-points.
        de_for_qpe: epsilon for phase estimation.
    """
    return np.ceil(5.652 + np.log2(num_spin_orbs * num_kpts / de_for_qpe))


def compute_cost(
        num_spin_orbs: int,
        lambda_tot: float,
        num_aux: int,
        num_eig: int,
        kmesh: List[int],
        dE_for_qpe: float = 0.0016,
        chi: int = 10,
        beta: Union[int, None] = None,
) -> ResourceEstimates:
    """Determine fault-tolerant costs using double factorized Hamiltonian.

    Light wrapper around _compute_cost to automate choice of steps parameter.

    Arguments:
        num_spin_orbs: the number of spin-orbitals
        lambda_tot: the lambda-value for the Hamiltonian
        num_aux: number auxiliary vectors for the first factorization.
        num_eig: number of retained eigenvalues for the second factorization.
        dE_for_qpe: allowable error in phase estimation
        chi: the number of bits for the representation of the coefficients
        beta: the number of bits for controlled rotations
    Returns:
        resources: double factorized resource estimates
    """
    # run once to determine steps parameter
    if beta is None:
        num_kpts = int(np.prod(kmesh))
        beta = compute_beta_for_resources(num_spin_orbs, num_kpts, dE_for_qpe)
    init_cost = _compute_cost(
        n=num_spin_orbs,
        lam=lambda_tot,
        dE=dE_for_qpe,
        L=num_aux,
        Lxi=num_eig,
        chi=chi,
        beta=beta,
        Nkx=kmesh[0],
        Nky=kmesh[1],
        Nkz=kmesh[2],
        steps=20_000,
    )
    steps = init_cost[0]
    final_cost = _compute_cost(
        n=num_spin_orbs,
        lam=lambda_tot,
        dE=dE_for_qpe,
        L=num_aux,
        Lxi=num_eig,
        chi=chi,
        beta=beta,
        Nkx=kmesh[0],
        Nky=kmesh[1],
        Nkz=kmesh[2],
        steps=steps,
    )
    estimates = ResourceEstimates(
        toffolis_per_step=final_cost[0],
        total_toffolis=final_cost[1],
        logical_qubits=final_cost[2],
    )
    return estimates


def _compute_cost(
        n: int,
        lam: float,
        dE: float,
        L: int,
        Lxi: int,
        chi: int,
        beta: int,
        Nkx: int,
        Nky: int,
        Nkz: int,
        steps: int,
        verbose: bool = False,
) -> Tuple[int, int, int]:
    """Determine fault-tolerant costs using DF decomposition.

    Arguments:
        n: the number of spin-orbitals.
        lam: the lambda-value for the Hamiltonian
        dE: allowable error in phase estimation
        L: the rank of the first decomposition
        Lxi: the total number of eigenvectors
        chi: the number of bits for the representation of the coefficients
        beta: the number of bits for the rotations
        Nkx: Number of k-points in x-direction
        Nky: Number of k-points in y-direction
        Nkz:  Number of k-points in z-direction
        steps: an approximate number of steps to choose the precision of
            single qubit rotations in preparation of the equal superposition
            state.
        verbose: do additional printing of intermediates?

    Returns:
        step_cost: Toffolis per step
        total_cost: Total number of Toffolis
        ancilla_cost: Total ancilla cost
    """
    nNk = (max(np.ceil(np.log2(Nkx)), 1) + max(np.ceil(np.log2(Nky)), 1) +
           max(np.ceil(np.log2(Nkz)), 1))
    Nk = Nkx * Nky * Nkz

    # The number of bits used for the second register.
    nxi = np.ceil(np.log2(Nk * n // 2))

    # The number of bits for the contiguous register.
    nLxi = np.ceil(np.log2(Lxi + Nk * n // 2))

    # The number of bits used for the first register.
    nL = np.ceil(np.log2(L + 1))

    # The power of 2 that is a factor of L + 1
    if L % 2 == 1:
        factors = factorint(L + 1)
        eta = factors[min(list(sorted(factors.keys())))]
    else:
        eta = 0

    oh = [0] * 20
    for p in range(20):
        v = np.round(
            np.power(2, p + 1) / (2 * np.pi) *
            arccos(np.power(2, nL) / np.sqrt((L + 1) / 2**eta) / 2))
        oh[p] = np.real(steps * (1 / (np.sin(3 * arcsin(
            np.cos(v * 2 * np.pi / np.power(2, p + 1)) * np.sqrt(
                (L + 1) / 2**eta) / np.power(2, nL)))**2) - 1) + 4 * (p + 1))

    # Bits of precision for rotation
    br = int(np.argmin(oh) + 1)

    # The following costs are from the list starting on page 50.

    # The cost for preparing an equal superposition for preparing the first
    # register in step 1 (a). We double this cost to account for the inverse.
    cost1a = 2 * (3 * nL + 2 * br - 3 * eta - 9)

    # The output size for the QROM for the first state preparation in Eq. (C27)
    bp1 = nL + chi

    # The cost of the QROM for the first state preparation in step 1 (b) and
    # its inverse.
    cost1b = QR3(L + 1, bp1)[1] + QI(L + 1)[1]

    # The cost for the inequality test, controlled swap and their inverse in
    # steps 1 (c) and (d)
    cost1cd = 2 * (chi + nL)

    # The total cost for preparing the first register in step 1.
    cost1 = cost1a + cost1b + cost1cd

    # The output size for the QROM for the data to prepare the equal
    # superposition on the second register, as given in Eq. (C29).
    bo = nxi + nLxi + br + 1
    bo = bo + nNk  # account for k-point sampling outputting values of Q.

    # This is step 2. This is the cost of outputting the data to prepare the
    # equal superposition on the second register. We will assume it is not
    # uncomputed, because we want to keep the offset for applying the QROM for
    # outputting the rotations.
    cost2 = QR3(L + 1, bo)[1] + QI(L + 1)[1]

    # The number of bits for rotating the ancilla for the second preparation.
    # We are just entering this manually because it is a typical value.
    br = 7

    # The cost of preparing an equal superposition over the second register in
    # a controlled way. We pay this cost 4 times.
    cost3a = 4 * (7 * nxi + 2 * br - 6)

    # The cost of the offset to apply the QROM for state preparation on the
    # second register.
    cost3b = 4 * (nLxi - 1)

    bp2 = nxi + chi + 2

    # The cost of the QROMs and inverse QROMs for the state preparation, where
    # in the first one we need + n/2 to account for the one-electron terms.
    cost3c = (QR3(Lxi + Nk * n // 2, bp2)[1] + QI(Lxi + Nk * n // 2)[1] +
              QR3(Lxi, bp2)[1] + QI(Lxi)[1])

    # The inequality test and state preparations.
    cost3d = 4 * (nxi + chi)

    # The total costs for state preparations on register 2.
    cost3 = cost3a + cost3b + cost3c + cost3d

    # The cost of adding offsets in steps 4 (a) and (h).
    cost4ah = 4 * (nLxi - 1)

    # The costs of the QROMs and their inverses in steps 4 (b) and (g).
    cost4bg = (QR3(Lxi + Nk * n // 2, 2 * n * beta + nNk)[1] +
               QI(Lxi + Nk * n // 2)[1] + QR3(Lxi, 4 * n)[1] + QI(Lxi)[1])

    # The cost of the controlled swaps based on the spin qubit in steps 4c and f
    cost4cf = 2 * n * Nk

    # The controlled rotations in steps 4 (d) and (f).
    cost4df = 16 * n * (beta - 2)

    # The controlled Z operations in the middle for step 4 (e).
    cost4e = 3
    cost4e = (
        cost4e + 4 * n * Nk + 12 * nNk
    )  # extra cost of swapping into working registers and computing k-Q.

    # (*Adjustment for modular arithmetic costs.*)
    if Nkx == 2**np.ceil(np.log2(Nkx)):
        cost4e = cost4e - 2 * np.ceil(np.log2(Nkx))
    if Nky == 2**np.ceil(np.log2(Nky)):
        cost4e = cost4e - 2 * np.ceil(np.log2(Nky))
    if Nkz == 2**np.ceil(np.log2(Nkz)):
        cost4e = cost4e - 2 * np.ceil(np.log2(Nkz))

    # This is the cost of the controlled rotations for step 4.
    cost4 = cost4ah + cost4bg + cost4cf + cost4df + cost4e

    # This is the cost of the reflection on the second register from step 6.
    cost6 = nxi + chi + 2

    # The cost of the final reflection req'd to construct the step of the
    # quantum walk from step 9.
    cost9 = nL + nxi + chi + 1

    # The extra two qubits for unary iteration and making the rflxn controlled.
    cost10 = 2

    # The Toffoli cost for a single step
    cost = cost1 + cost2 + cost3 + cost4 + cost6 + cost9 + cost10

    # The number of steps needed
    iters = np.ceil(np.pi * lam / (2 * dE))

    # Now the number of qubits from the list on page 54.

    k1 = np.power(2, QR3(Lxi + Nk * n // 2, 2 * n * beta + nNk)[0])

    # The control register for phase estimation and iteration on it.
    ac1 = np.ceil(np.log2(iters + 1)) * 2 - 1

    # The system qubits
    ac2 = n * Nk

    # The first register prepared, a rotated qubit and a flag qubit.
    ac3 = nL + 2

    # The output of the QROM, the equal superposition state and a flag qubit.
    ac4 = nL + chi * 2 + 1

    # The data used for preparing the equal superposition state on the 2nd reg
    ac5 = bo

    # The second register, a rotated qubit and a flag qubit.
    ac6 = nxi + 2

    # The second preparation QROM output.
    ac8 = bp2

    # The equal superposition state and the result of the inequality test.
    ac9 = chi + 1

    # The angles for rotations.
    ac10 = k1 * (2 * n * beta + nNk)

    # The phase gradient state.
    ac11 = beta

    # A control qubit for the spin.
    ac12 = 1

    # A T state.
    ac13 = 1

    # extra for kpoint
    ac14 = nNk + 2 * n

    if verbose:
        print("[*] Top of routine")
        print("  [+] nxi = ", nxi)
        print("  [+] nLxi = ", nLxi)
        print("  [+] nL = ", nL)
        print("  [+] eta = ", eta)
        print("  [+] cost3 = ", cost3)
        print("  [+] cost4 = ", cost4)
        print("  [+] cost = ", cost)
        print("  [+] iters = ", iters)

    ancilla_cost = (ac1 + ac2 + ac3 + ac4 + ac5 + ac6 + ac8 + ac9 + ac10 +
                    ac11 + ac12 + ac13 + ac14)

    # Sanity checks before returning as int
    assert cost.is_integer()
    assert iters.is_integer()
    assert ancilla_cost.is_integer()

    step_cost = int(cost)
    total_cost = int(cost * iters)
    ancilla_cost = int(ancilla_cost)

    return step_cost, total_cost, ancilla_cost



#coverage:ignore
""" Utilities for FT costing calculations """
from typing import Tuple, Optional
import sys
import os
import h5py
import numpy as np


def QR(L: int, M1: int) -> Tuple[int, int]:
    """ This gives the optimal k and minimum cost for a QROM over L values of
        size M.

    Args:
        L (int) -
        M1 (int) -

    Returns:
       k_opt (int) - k that yields minimal (optimal) cost of QROM
       val_opt (int) - minimal (optimal) cost of QROM
    """
    k = 0.5 * np.log2(L / M1)
    try:
        assert k >= 0
    except AssertionError:
        sys.exit("In function QR: L is smaller than M: increase RANK or lower "
                 "THRESH (or alternatively decrease CHI)")
    value = lambda k: L / np.power(2.0, k) + M1 * (np.power(2.0, k) - 1)
    k_int = [np.floor(k), np.ceil(k)]  # restrict optimal k to integers
    k_opt = k_int[np.argmin(value(k_int))]  # obtain optimal k
    val_opt = np.ceil(value(k_opt))  # obtain ceiling of optimal value given k
    assert k_opt.is_integer()
    assert val_opt.is_integer()
    return int(k_opt), int(val_opt)


def QR2(L1: int, L2: int, M1: int) -> Tuple[int, int, int]:
    """ This gives the optimal k values and minimum cost for a QROM using
        two L values of size M,
        e.g. the optimal k values for the QROM on two registers.
    Args:
        L1 (int) -
        L2 (int) -
        M1 (int) -

    Returns:
       k1_opt (int) - k1 that yields minimal (optimal) cost of QROM with two reg
       k2_opt (int) - k2 that yields minimal (optimal) cost of QROM with two reg
       val_opt (int) - minimal (optimal) cost of QROM
    """

    k1_opt, k2_opt = 0, 0
    val_opt = 1e50
    # Doing this as a stupid loop for now, worth refactoring but cost is quick
    # Biggest concern is if k1 / k2 range is not large enough!
    for k1 in range(1, 17):
        for k2 in range(1, 17):
            value = np.ceil(L1 / np.power(2.0, k1)) * np.ceil(L2 / \
                np.power(2.0, k2)) +\
                M1 * (np.power(2.0, k1 + k2) - 1)
            if value < val_opt:
                val_opt = value
                k1_opt = k1
                k2_opt = k2

    assert val_opt.is_integer()
    return int(np.power(2, k1_opt)), int(np.power(2, k2_opt)), int(val_opt)


def QI(L: int) -> Tuple[int, int]:
    """ This gives the opt k and minimum cost for an inverse QROM over L vals

    Args:
        L (int) -

    Returns:
       k_opt (int) - k that yiles minimal (optimal) cost of inverse QROM
       val_opt (int) - minimal (optimal) cost of inverse QROM
    """
    k = 0.5 * np.log2(L)
    assert k >= 0
    value = lambda k: L / np.power(2.0, k) + np.power(2.0, k)
    k_int = [np.floor(k), np.ceil(k)]  # restrict optimal k to integers
    k_opt = k_int[np.argmin(value(k_int))]  # obtain optimal k
    val_opt = np.ceil(value(k_opt))  # obtain ceiling of optimal value given k
    assert k_opt.is_integer()
    assert val_opt.is_integer()
    return int(k_opt), int(val_opt)


# Is this ever used? It's defined in costingsf.nb, but I don't it's ever called.
def QI2(L1: int, L2: int) -> Tuple[int, int, int]:
    """ This gives the optimal k values and minimum cost for inverse QROM
        using two L values,
        e.g. the optimal k values for the inverse QROM on two registers.

    Args:
        L1 (int) -
        L2 (int) -

    Returns:
       k1_opt (int) - k1 with minimal (optimal) cost of inverse QROM with 2 regs
       k2_opt (int) - k2 with minimal (optimal) cost of inverse QROM with 2 regs
       val_opt (int) - minimal (optimal) cost of inverse QROM with two registers
    """

    k1_opt, k2_opt = 0, 0
    val_opt = 1e50
    # Doing this as a stupid loop for now, worth refactoring but cost is quick
    # Biggest concern is if k1 / k2 range is not large enough!
    for k1 in range(1, 17):
        for k2 in range(1, 17):
            value = np.ceil(L1 / np.power(2.0, k1)) * np.ceil(L2 / \
                np.power(2.0, k2)) +\
                np.power(2.0, k1 + k2)
            if value < val_opt:
                val_opt = value
                k1_opt = k1
                k2_opt = k2

    assert val_opt.is_integer()
    return int(np.power(2, k1_opt)), int(np.power(2, k2_opt)), int(val_opt)


def power_two(m: int) -> int:
    """ Return the power of two that is a factor of m """
    assert m >= 0
    if m % 2 == 0:
        count = 0
        while (m > 0) and (m % 2 == 0):
            m = m // 2
            count += 1
        return count
    return 0


class RunSilent(object):
    """ Context manager to prevent function writing to stdout/stderr
        e.g. for noisy_function(), wrap it like so

        with RunSilent():
            noisy_function()

        ... and your terminal will no longer be littered with prints
    """

    def __init__(self, stdout=None, stderr=None):
        self.devnull = open(os.devnull, 'w')
        self._stdout = stdout or self.devnull or sys.stdout
        self._stderr = stderr or self.devnull or sys.stderr

    def __enter__(self):
        self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
        self.old_stdout.flush()
        self.old_stderr.flush()
        sys.stdout, sys.stderr = self._stdout, self._stderr

    def __exit__(self, exc_type, exc_value, traceback):
        self._stdout.flush()
        self._stderr.flush()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr
        self.devnull.close()


def eigendecomp(M, tol=1.15E-16):
    """ Decompose matrix M into L.L^T where rank(L) < rank(M) to some threshold

    Args:
       M (np.ndarray) - (N x N) positive semi-definite matrix to be decomposed
       tol (float) - eigenpairs with eigenvalue above tol will be kept

    Returns:
       L (np.ndarray) - (K x N) array such that K <= N and L.L^T = M
    """
    eigenvalues, eigenvectors = np.linalg.eigh(M)

    # Put in descending order
    eigenvalues = eigenvalues[::-1]
    eigenvectors = eigenvectors[:, ::-1]

    # Truncate
    idx = np.where(eigenvalues > tol)[0]
    eigenvalues, eigenvectors = eigenvalues[idx], eigenvectors[:, idx]

    # eliminate eigenvalues from eigendecomposition
    L = np.einsum("ij,j->ij", eigenvectors, np.sqrt(eigenvalues))

    return L



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import numpy.typing as npt
import pandas as pd

from pyscf.pbc import scf
from pyscf.pbc.tools.k2gamma import kpts_to_kmesh

from openfermion.resource_estimates.pbc.hamiltonian import (
    build_hamiltonian,)
from openfermion.resource_estimates.pbc.hamiltonian.cc_extensions import (
    build_approximate_eris,
    build_cc_inst,
    build_approximate_eris_rohf,
)
from openfermion.resource_estimates.pbc.sf.compute_lambda_sf import (
    compute_lambda,)
from openfermion.resource_estimates.pbc.sf.compute_sf_resources import (
    compute_cost,)
from openfermion.resource_estimates.pbc.sf.sf_integrals import (
    SingleFactorization,)
from openfermion.resource_estimates.pbc.resources import PBCResources


def generate_costing_table(
        pyscf_mf: scf.HF,
        naux_cutoffs: npt.NDArray[np.int32],
        name: str = "pbc",
        chi: int = 10,
        dE_for_qpe=0.0016,
        energy_method="MP2",
) -> pd.DataFrame:
    """Generate resource estimate costing table given a set of cutoffs for
        single-factorized Hamiltonian.

    Arguments:
        pyscf_mf: k-point pyscf mean-field object
        naux_cutoffs: Array of (integer) auxiliary index cutoff values
        name: Optional descriptive name for simulation.
        chi: the number of bits for the representation of the coefficients
        dE_for_qpe: Phase estimation epsilon.

    Returns
        resources: Table of resource estimates.
    """
    kmesh = kpts_to_kmesh(pyscf_mf.cell, pyscf_mf.kpts)
    cc_inst = build_cc_inst(pyscf_mf)
    exact_eris = cc_inst.ao2mo()
    if energy_method == "MP2":
        energy_function = lambda x: cc_inst.init_amps(x)
        reference_energy, _, _ = energy_function(exact_eris)
    elif energy_method == "CCSD":
        energy_function = lambda x: cc_inst.kernel(eris=x)
        reference_energy, _, _ = energy_function(exact_eris)
    else:
        raise ValueError(f"Unknown value for energy_method: {energy_method}")

    hcore, chol = build_hamiltonian(pyscf_mf)
    num_spin_orbs = 2 * hcore[0].shape[-1]
    num_kpts = np.prod(kmesh)

    sf_resource_obj = PBCResources(
        system_name=name,
        num_spin_orbitals=num_spin_orbs,
        num_kpts=num_kpts,
        dE=dE_for_qpe,
        chi=chi,
        energy_method=energy_method,
        exact_energy=np.real(reference_energy),
    )
    approx_eris = exact_eris
    for cutoff in naux_cutoffs:
        sf_helper = SingleFactorization(cholesky_factor=chol,
                                        kmf=pyscf_mf,
                                        naux=cutoff)
        if pyscf_mf.cell.spin == 0:
            approx_eris = build_approximate_eris(cc_inst,
                                                 sf_helper,
                                                 eris=approx_eris)
        else:
            approx_eris = build_approximate_eris_rohf(cc_inst,
                                                      sf_helper,
                                                      eris=approx_eris)
        approx_energy, _, _ = energy_function(approx_eris)

        sf_lambda = compute_lambda(hcore, sf_helper)
        sparse_res_cost = compute_cost(
            num_spin_orbs,
            sf_lambda.lambda_total,
            sf_lambda.num_aux,
            list(kmesh),
            chi=chi,
            dE_for_qpe=dE_for_qpe,
        )

        sf_resource_obj.add_resources(
            ham_properties=sf_lambda,
            resource_estimates=sparse_res_cost,
            cutoff=sf_lambda.num_aux,
            approx_energy=np.real(approx_energy),
        )

    return sf_resource_obj.to_dataframe()



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.thc.\
        generate_costing_table_thc import generate_costing_table
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_generate_costing_table_thc():
    mf = make_diamond_113_szv()
    thc_rank_params = np.array([2, 4, 6])
    table = generate_costing_table(
        mf,
        thc_rank_params=thc_rank_params,
        chi=10,
        beta=22,
        dE_for_qpe=1e-3,
        bfgs_maxiter=10,
        adagrad_maxiter=10,
        fft_df_mesh=[11] * 3,
    )
    assert np.allclose(table.dE, 1e-3)
    assert np.allclose(table.chi, 10)
    assert np.allclose(table.beta, 22)
    assert np.allclose(table.cutoff, thc_rank_params)



#coverage:ignore
""" Double factorization rank reduction of ERIs """
import numpy as np
from openfermion.resource_estimates.utils import eigendecomp


def factorize(eri_full, thresh):
    """ Do double factorization of the ERI tensor

    Args:
       eri_full (np.ndarray) - 4D (N x N x N x N) full ERI tensor
       thresh (float) - threshold for double factorization

    Returns:
       eri_rr (np.ndarray) - 4D approximate ERI tensor reconstructed
           from df_factors vectors
       df_factors (np.ndarray) - 3D (N x N x M) tensor containing DF vectors
       rank (int) - rank retained from initial eigendecomposition
       num_eigenvectors (int) - number of eigenvectors
    """

    n_orb = eri_full.shape[0]
    assert n_orb**4 == len(eri_full.flatten())

    # First, do an eigendecomposition of ERIs
    L = eigendecomp(eri_full.reshape(n_orb**2, n_orb**2), tol=0.0)
    L = L.reshape(n_orb, n_orb, -1)  # back to (N x N x rank)

    sf_rank = L.shape[2]

    num_eigenvectors = 0  # rolling number of eigenvectors
    df_factors = []  # collect the selected vectors
    for rank in range(sf_rank):
        Lij = L[:, :, rank]
        e, v = np.linalg.eigh(Lij)
        normSC = np.sum(np.abs(e))

        truncation = normSC * np.abs(e)

        idx = truncation > thresh
        plus = np.sum(idx)
        num_eigenvectors += plus

        if plus == 0:
            break

        e_selected = np.diag(e[idx])
        v_selected = v[:, idx]

        Lij_selected = v_selected.dot(e_selected).dot(v_selected.T)
        df_factors.append(Lij_selected)

    # raw factors from DF algorithm
    df_factors = np.asarray(df_factors).T

    # double-factorized and re-constructed ERIs
    eri_rr = np.einsum('ijP,klP', df_factors, df_factors, optimize=True)

    return eri_rr, df_factors, rank, num_eigenvectors



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module for performing ISDF THC factorization of k-point dependent integrals.

The ISDF (interpolative separable density fitting) implementation currently
provides a tensor hypercontraction (THC)-like factorization of the two electron
integrals which should converge to the exact representation (fast fourier
transform density fitting (FFTDF)) of the ERIs in the limit of large THC rank.
This differs from the assumption of using range-separated gaussian density
fitting (RSGDF) throughout the rest of the resource estimation scripts. However,
we typically are only interested in ISDF as an initial guess for the THC factors
which are then subsequently reoptimized to regularize lambda. The assumption
here is that FFTDF / ISDF is a good enough approximation to the RSGDF ERIs and
thus serves as a good initial guess.
"""

import itertools
from dataclasses import dataclass
from typing import Tuple, Union

import numpy as np
import numpy.typing as npt
import scipy.linalg

from pyscf.pbc import df, gto, scf, tools
from pyscf.pbc.dft import numint
from pyscf.pbc.dft.gen_grid import UniformGrids
from pyscf.pbc.lib.kpts_helper import conj_mapping, get_kconserv, unique

from openfermion.resource_estimates.pbc.thc.factorizations.kmeans import (
    KMeansCVT)
from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)


def check_isdf_solution(
        orbitals: npt.NDArray,
        interp_orbitals: npt.NDArray,
        xi: npt.NDArray,
) -> float:
    r"""Check accuracy of isdf least squares solution.

    Very costly and should only be used for testing purposes.

    Args:
        orbitals: Orbitals on full real space grid. [num_grd, num_orb]
        interp_orbitals: interpolating orbitals (those orbitals evaluated on
            interpolating points.) [num_interp, num_orb]
        xi: interpolating vectors. [num_grid, num_interp]
    Returns:
        error: |phi_{ij}(r) - \sum_m xi_m(r) phi_{ij}(r_m)
        orbitals: npt.NDArray:
        interp_orbitals: npt.NDArray:
        xi: npt.NDArray:
    """

    lhs = np.einsum("Ri,Rj->Rij", orbitals.conj(), orbitals, optimize=True)
    rhs = np.einsum("mi,mj->mij",
                    interp_orbitals.conj(),
                    interp_orbitals,
                    optimize=True)
    lhs_check = np.einsum("Rm,mij->Rij", xi, rhs, optimize=True)
    return np.linalg.norm(lhs - lhs_check)


def solve_isdf(orbitals: npt.NDArray,
               interp_indx: npt.NDArray) -> Tuple[npt.NDArray, npt.NDArray]:
    """Solve for interpolating vectors given interpolating points and orbitals.

    Used for both supercell and k-point ISDF factorizations.

    Args:
        orbitals: orbitals on a grid of shape [num_grid_points, num_orbitals]
        interp_indx: array indexing interpolating points (subset of grid
            points to use selected by K-Means algorithm. shape is
            [num_interp_points].
        orbitals: npt.NDArray:
        interp_indx: npt.NDArray:

    Returns:
        tuple: (Interpolang vectors, interpolating orbitals) (xi_mu(r),
            phi_i(r_mu)). Note xi_mu(r) is called theta[R, mu] in keeping with
            original ISDF notation.
    """
    interp_orbitals = orbitals[interp_indx]
    # Form pseudo-densities
    # P[R, r_mu] = \sum_{i} phi_{R,i}^* phi_{mu,i}
    pseudo_density = np.einsum("Ri,mi->Rm",
                               orbitals.conj(),
                               interp_orbitals,
                               optimize=True)
    # [Z C^]_{J, mu} = (sum_i phi_{i, J}^* phi_{i, mu}) (sum_j phi_{j, J}
    # phi_{i, mu})
    zc_dag = np.einsum("Rm,Rm->Rm",
                       pseudo_density,
                       pseudo_density.conj(),
                       optimize=True)
    # Just down sample from ZC_dag
    cc_dag = zc_dag[interp_indx].copy()
    # Solve ZC_dag = Theta CC_dag
    # Theta = ZC_dag @ CC_dag_inv
    # Solve ZC_dag = Theta CC_dag
    # -> ZC_dag^T = CC_dag^T Theta^T
    # rcond = None uses MACH_EPS * max(M,N) for least squares convergence.
    theta_dag, _, _, _ = np.linalg.lstsq(cc_dag.conj().T,
                                         zc_dag.conj().T,
                                         rcond=None)
    return theta_dag.conj().T, interp_orbitals


def supercell_isdf(
        mydf: df.FFTDF,
        interp_indx: npt.NDArray,
        orbitals: npt.NDArray,
        grid_points: npt.NDArray,
        kpoint=np.zeros(3),
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
    r"""
    Build ISDF-THC tensors.

    Given the orbitals evaluated on a (dense) real space grid, and a set of
    interpolating points (indexed  by interp_indx) determine the interpolating
    orbitals (chi), central tensor (zeta), and interpolating vectors Theta (also
    called xi).

    Args:
        mydf: instance of pyscf.pbc.df.FFTDF object.
        interp_indx: array indexing interpolating points determined through
            K-Means CVT procedure. Dimension [num_interp_points]
        orbitals: orbitals on a grid of shape [num_grid_points, num_orbitals]
        grid_points: Real space grid. Dimension [num_grid_points, num_dim],
            num_dim is 1, 2 or 3 for 1D, 2D, 3D.

    Returns:
        tuple: (chi, zeta, Theta): orbitals on interpolating
            points, zeta (central tensor), and matrix of interpolating vectors
            Theta of dimension [num_grid_points, num_interp_points] (also called
            xi_mu(r)), where num_grid_points is the number of real space grid
            points and num_interp_points is the number of interpolating points.
    """

    cell = mydf.cell
    num_grid_points = len(grid_points)

    theta, chi = solve_isdf(orbitals, interp_indx)

    # FFT Theta[R, mu] -> Theta[mu, G]
    # Transpose as fft expects contiguous.
    theta_g = tools.fft(theta.T, mydf.mesh)
    coulg = tools.get_coulG(cell, k=kpoint, mesh=mydf.mesh)
    weighted_coulg = coulg * cell.vol / num_grid_points**2.0

    # zeta_{mu,nu} = \sum_G 4pi/(omega * G^2) zeta_{mu,G} * (zeta_G*){nu, G}
    theta_g_tilde = np.einsum("iG,G->iG", theta_g, weighted_coulg)
    zeta = (theta_g_tilde) @ theta_g.conj().T
    return chi, zeta, theta


def build_kpoint_zeta(
        df_inst: df.FFTDF,
        q: int,
        delta_g: npt.NDArray,
        delta_g_prime: npt.NDArray,
        grid_points: npt.NDArray,
        xi_mu: npt.NDArray,
) -> npt.NDArray:
    """Build k-point THC zeta (central tensor).

    Built for for given q, delta_g, delta_g_prime.

    Args:
        mydf: instance of pyscf.pbc.df.FFTDF object.
        q: Momentum transfer (in 1BZ).
        delta_g: Reciprocal lattice vector satisfying Q - (Q-k) = delta_G
        delta_g_prime: Reciprocal lattice vector satisfying Q - (Q-k') = delta_G
        grid_points: Real space grid. Dimension [num_grid_points, num_dim],
            num_dim is 1, 2 or 3 for 1D, 2D, 3D.
        xi_mu: array containing interpolating vectors determined during ISDF
            procedure

    Returns:
        zeta: central tensor of dimension [num_interp_points, num_interp_points]

    """
    cell = df_inst.cell
    num_grid_points = grid_points.shape[0]
    # delta_G - delta_G_prime because we have Gpq and Gsr and Gsr = -Grs, phase
    # = Delta G = Gpq + Grs
    phase_factor = np.exp(-1j * (np.einsum(
        "x,Rx->R", delta_g - delta_g_prime, grid_points, optimize=True)))
    # Minus sign again due to we use Q = kp - kq, but we should have
    # V(G + k_q - k_p)
    coulg = tools.get_coulG(cell, k=-(q + delta_g), mesh=df_inst.mesh)
    weighted_coulg = coulg * cell.vol / num_grid_points
    xi_mug = tools.fft(xi_mu.T, df_inst.mesh)
    xi_mug *= weighted_coulg
    vr = tools.ifft(xi_mug, df_inst.mesh)
    zeta = np.einsum("R,Rn,mR->mn", phase_factor, xi_mu, vr, optimize=True)
    return zeta


def build_kpoint_zeta_single_tranlsation(
        df_inst: df.FFTDF,
        q: int,
        delta_g: npt.NDArray,
        grid_points: npt.NDArray,
        xi_mu: npt.NDArray,
) -> npt.NDArray:
    """Build k-point THC zeta (central tensor)

    Built for for given q, delta_g. Note single delta_g (single translation
    vector) in controast to the _double_translation variant above.

    Args:
      mydf: instance of pyscf.pbc.df.FFTDF object.
      q: Momentum transfer kp-kq.
      delta_G: Reciprocal lattice vector satisfying Q - (Q-k) = delta_G
      grid_points: Real space grid. Dimension [num_grid_points, num_dim],
        num_dim is 1, 2 or 3 for 1D, 2D, 3D.
      xi_mu: array containing interpolating vectors determined during ISDF
        procedure

    Returns:
      zeta: central tensor of dimension [num_interp_points, num_interp_points]

    """
    cell = df_inst.cell
    num_grid_points = grid_points.shape[0]
    # delta_G - delta_G_prime because we have Gpq and Gsr and Gsr = -Grs, phase
    # = Delta G = Gpq + Grs
    phase_factor = np.exp(
        -1j * (np.einsum("x,Rx->R", delta_g, grid_points, optimize=True)))
    # Minus sign again due to we use Q = kp - kq, but we should have
    # V(G + k_q - k_p)
    coulg = tools.get_coulG(cell, k=-q, mesh=df_inst.mesh)
    weighted_coulg = coulg * cell.vol / num_grid_points
    xi_mug = tools.fft(xi_mu.T, df_inst.mesh)
    xi_mug *= weighted_coulg
    vr = tools.ifft(xi_mug, df_inst.mesh)
    zeta = np.einsum("R,Rn,mR->mn", phase_factor, xi_mu, vr, optimize=True)
    return zeta


def build_g_vectors(cell: gto.Cell) -> npt.NDArray:
    """Build all 27 Gvectors

    Args:
        cell: pyscf.pbc.gto.Cell object.
        cell: gto.Cell:

    Returns:
        tuple: g_dict a dictionary mapping miller index to appropriate
            g_vector index and g_vectors array of 27 g_vectors shape [27, 3].

    """
    g_dict = {}
    g_vectors = np.zeros((27, 3), dtype=np.float64)
    lattice_vectors = cell.lattice_vectors()
    indx = 0
    for n1, n2, n3 in itertools.product(range(-1, 2), repeat=3):
        g_dict[(n1, n2, n3)] = indx
        g_vectors[indx] = np.einsum("n,ng->g", (n1, n2, n3),
                                    cell.reciprocal_vectors())
        miller_indx = np.rint(
            np.einsum("nx,x->n", lattice_vectors, g_vectors[indx]) /
            (2 * np.pi))
        assert (miller_indx == (n1, n2, n3)).all()
        indx += 1
    return g_dict, g_vectors


def find_unique_g_vectors(g_vectors: npt.NDArray, g_mapping: npt.NDArray
                         ) -> Tuple[npt.NDArray, npt.NDArray]:
    """Find all unique G-vectors and build mapping to original set.

    Args:
        g_vectors: array of 27 g-vectors.
        g_mapping: array of 27 g-vectors.

    Returns:
      unique_g_index in range [0,...,num_unique_gs[iq]], and delta_gs are the
        unique g-vectors of size [num_qpoints, num_unique_gs[iq]].

    """
    unique_mapping = np.zeros_like(g_mapping)
    num_qpoints = g_mapping.shape[0]
    delta_gs = np.zeros((num_qpoints,), dtype=object)
    for iq in range(num_qpoints):
        unique_g = np.unique(g_mapping[iq])
        delta_gs[iq] = g_vectors[unique_g]
        # build map to unique index
        unique_mapping[iq] = [
            ix for el in g_mapping[iq] for ix in np.where(unique_g == el)[0]
        ]

    return unique_mapping, delta_gs


def build_g_vector_mappings_double_translation(
        cell: gto.Cell,
        kpts: npt.NDArray,
        momentum_map: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
    """build g-vector mappings that map k-point differences to 1bz.

    args:
      cell: pyscf.pbc.gto.cell object.
      kpts: array of kpoints.
      momentum_map: momentum mapping to satisfy q = (k_p - k_q) mod g.
        momentum_map[iq, ikp] = ikq.

    returns:
      tuple: (g_vectors, gpq_mapping, gpq_mapping_unique, delta_gs), g_vectors
        is a list of all 27
      g-vectors and gpq_mapping[iq, kp] = indx_gpq, where gpq = kpts[ikp] -
      kpts[ikq] - kpts[iq], i.e. returns index to g-vector (consistent with
      g_vectors) satisfying this condition. gpq_mapping_unique provides
      mapping to unique g_vector index. delta_gs provides compressed lists of
      unique g vectors.

    """
    g_dict, g_vectors = build_g_vectors(cell)
    lattice_vectors = cell.lattice_vectors()
    num_kpts = len(kpts)
    gpq_mapping = np.zeros((num_kpts, num_kpts), dtype=np.int32)
    num_kpts = len(kpts)
    for iq in range(num_kpts):
        for ikp in range(num_kpts):
            ikq = momentum_map[iq, ikp]
            delta_gpq = (kpts[ikp] - kpts[ikq]) - kpts[iq]
            miller_indx = np.rint(
                np.einsum("wx,x->w", lattice_vectors, delta_gpq) / (2 * np.pi))
            gpq_mapping[iq, ikp] = g_dict[tuple(miller_indx)]
    gpq_mapping_unique, delta_gs = find_unique_g_vectors(g_vectors, gpq_mapping)
    return g_vectors, gpq_mapping, gpq_mapping_unique, delta_gs


def get_miller(lattice_vectors: npt.NDArray, g: npt.NDArray) -> npt.NDArray:
    """convert g to miller indx.

    args:
      lattice_vectors: array of lattice vectors.
      g: reciprocal lattice vector.

    returns:
      miller_index: 3d array of miller indices.

    """
    miller_indx = np.rint(
        np.einsum("wx,x->w", lattice_vectors, g) / (2 * np.pi)).astype(np.int32)
    return miller_indx


def build_minus_q_g_mapping(cell: gto.Cell, kpts: npt.NDArray,
                            momentum_map: npt.NDArray) -> npt.NDArray:
    """Build conjugat g map

    Mapping satisfies (-Q) + G + (Q + Gpq) = 0 (*)

    Args:
        cell: pyscf.pbc.gto.Cell object.
        kpts: array of kpoints.
        momentum_map: momentum mapping to satisfy q = (k_p - k_q) mod g.
            momentum_map[iq, ikp] = ikq.

    Returns:
        minus_q_mapping: [indx_minus_q, k]-th element yields index for g
            satisfying (*) above, where indx_minus_q is given by indx_minus_q =
            minus_k[q], and minus_k = conj_mapping(cell, kpts).
        minus_q_mapping_unique: indexes the appropriate g vector given by
            delta_gs[indx_minus_q][indx] = g, where indx =
            minus_q_mapping_unique[indx_minus_q, k], and deltags is built by
            build_g_vector_mappings_double_translation.
    """
    (
        g_vecs,
        g_map,
        _,
        delta_gs,
    ) = build_g_vector_mappings_double_translation(cell, kpts, momentum_map)
    g_dict, _ = build_g_vectors(cell)
    num_kpts = len(kpts)
    lattice_vectors = cell.lattice_vectors()
    minus_k_map = conj_mapping(cell, kpts)
    minus_q_mapping = np.zeros((num_kpts, num_kpts), dtype=np.int32)
    minus_q_mapping_unique = np.zeros((num_kpts, num_kpts), dtype=np.int32)
    for iq in range(num_kpts):
        minus_iq = minus_k_map[iq]
        for ik in range(num_kpts):
            gpq = g_vecs[g_map[iq, ik]]
            # complementary gpq (g in (*) in docstring)
            gpq_comp = -(kpts[minus_iq] + kpts[iq] + gpq)
            # find index in original set of 27
            igpq_comp = g_dict[tuple(get_miller(lattice_vectors, gpq_comp))]
            minus_q_mapping[minus_iq, ik] = igpq_comp
        indx_delta_gs = np.array([
            g_dict[tuple(get_miller(lattice_vectors, g))]
            for g in delta_gs[minus_iq]
        ])
        minus_q_mapping_unique[minus_iq] = [
            ix for el in minus_q_mapping[minus_iq]
            for ix in np.where(el == indx_delta_gs)[0]
        ]

    return minus_q_mapping, minus_q_mapping_unique


def build_g_vector_mappings_single_translation(
        cell: gto.Cell,
        kpts: npt.NDArray,
        kpts_pq: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
    """Build g-vector mappings that map k-point differences to 1BZ.

    Args:
      cell: pyscf.pbc.gto.Cell object.
      kpts: array of kpoints.
      kpts_pq: Unique list of kp - kq indices of shape [num_unique_pq, 2].

    Returns:
        G_vectors: list of all 27 G-vectors
        Gpqr_mapping: [iq, kr]-th element yields indx_Gpqr, where Gpqr =
            kpts[ikp] - kpts[ikq] + kpts[ikr] - kpts[iks], i.e. returns index to
            G_vectors (consistent with G_vectors) satisfying this condition.
        Gpqr_mapping_unique: provides mapping to unique G_vector index.
        Delta_gs: provides compressed lists of unique G vectors.

    """
    g_dict, g_vectors = build_g_vectors(cell)
    lattice_vectors = cell.lattice_vectors()
    num_kpts = len(kpts)
    gpqr_mapping = np.zeros((len(kpts_pq), num_kpts), dtype=np.int32)
    kconserv = get_kconserv(cell, kpts)
    for iq, (ikp, ikq) in enumerate(kpts_pq):
        q = kpts[ikp] - kpts[ikq]
        for ikr in range(num_kpts):
            iks = kconserv[ikp, ikq, ikr]
            delta_gpqr = q + kpts[ikr] - kpts[iks]
            miller_indx = np.rint(
                np.einsum("wx,x->w", lattice_vectors, delta_gpqr) / (2 * np.pi))
            gpqr_mapping[iq, ikr] = g_dict[tuple(miller_indx)]

    gpqr_mapping_unique, delta_gs = find_unique_g_vectors(
        g_vectors, gpqr_mapping)
    return g_vectors, gpqr_mapping, gpqr_mapping_unique, delta_gs


def inverse_g_map_double_translation(
        cell: gto.Cell,
        kpts: npt.NDArray,
        momentum_map: npt.NDArray,
) -> npt.NDArray:
    """for given q and g figure out all k which satisfy q - k + g = 0

    Args:
        cell: pyscf.pbc.gto.Cell object.
        kpts: array of kpoints.
        momentum_map: momentum mapping to satisfy Q = (k_p - k_q) mod G.
            momentum_map[iq, ikp] = ikq.

    Returns:
        inverse_map: ragged numpy array. inverse_map[iq, iG] returns array
            of size in range(0, num_kpts) and lists all k-point indices that
            satisfy G_pq[iq, ik] = iG, i.e. an array of all ik.
    """
    g_dict, g_vectors = build_g_vectors(cell)
    lattice_vectors = cell.lattice_vectors()
    num_kpts = len(kpts)
    gpq_mapping = np.zeros((num_kpts, num_kpts), dtype=np.int32)
    num_kpts = len(kpts)
    for iq in range(num_kpts):
        for ikp in range(num_kpts):
            ikq = momentum_map[iq, ikp]
            delta_gpq = (kpts[ikp] - kpts[ikq]) - kpts[iq]
            miller_indx = np.rint(
                np.einsum("wx,x->w", lattice_vectors, delta_gpq) / (2 * np.pi))
            gpq_mapping[iq, ikp] = g_dict[tuple(miller_indx)]

    inverse_map = np.zeros(
        (
            num_kpts,
            27,
        ),
        dtype=object,
    )
    for iq in range(num_kpts):
        for ig in range(len(g_vectors)):
            inverse_map[iq, ig] = np.array(
                [ik for ik in range(num_kpts) if gpq_mapping[iq, ik] == ig])

    return inverse_map


def build_eri_isdf_double_translation(
        chi: npt.NDArray,
        zeta: npt.NDArray,
        q_indx: int,
        kpts_indx: list,
        g_mapping: npt.NDArray,
) -> npt.NDArray:
    """Build (pkp qkq | rkr sks) from k-point ISDF factors.

    Args:
        chi: array of interpolating orbitals of shape
            [num_kpts, num_mo, num_interp_points]
        zeta: central tensor of dimension
            [num_kpts, num_G, num_G, num_interp_points, num_interp_points].
        q_indx: Index of momentum transfer.
        kpts_indx: List of kpt indices corresponding to [kp, kq, kr, ks]
        G_mapping: array to map kpts to G vectors [q_indx, kp] = G_pq

    Returns:
      eri:  (pkp qkq | rkr sks)
    """
    ikp, ikq, ikr, iks = kpts_indx
    gpq = g_mapping[q_indx, ikp]
    gsr = g_mapping[q_indx, iks]
    eri = np.einsum(
        "pm,qm,mn,rn,sn->pqrs",
        chi[ikp].conj(),
        chi[ikq],
        zeta[q_indx][gpq, gsr],
        chi[ikr].conj(),
        chi[iks],
        optimize=True,
    )
    return eri


def build_eri_isdf_single_translation(
        chi: npt.NDArray,
        zeta: npt.NDArray,
        q_indx: int,
        kpts_indx: list,
        g_mapping: npt.NDArray,
) -> npt.NDArray:
    """Build (pkp qkq | rkr sks) from k-point ISDF factors.

    Args:
        chi: array of interpolating orbitals of shape
            [num_kpts, num_mo, num_interp_points]
        zeta: central tensor of dimension
            [num_kpts, num_G, num_G, num_interp_points, num_interp_points].
        q_indx: Index of momentum transfer.
        kpts_indx: List of kpt indices corresponding to [kp, kq, kr, ks]
        G_mapping: array to map kpts to G vectors [q_indx, kp] = G_pq

    Returns:
        eri:  (pkp qkq | rkr sks)

    """
    ikp, ikq, ikr, iks = kpts_indx
    delta_g_indx = g_mapping[q_indx, ikr]
    eri = np.einsum(
        "pm,qm,mn,rn,sn->pqrs",
        chi[ikp].conj(),
        chi[ikq],
        zeta[q_indx][delta_g_indx],
        chi[ikr].conj(),
        chi[iks],
        optimize=True,
    )
    return eri


def kpoint_isdf_double_translation(
        df_inst: df.FFTDF,
        interp_indx: npt.NDArray,
        kpts: npt.NDArray,
        orbitals: npt.NDArray,
        grid_points: npt.NDArray,
        only_unique_g: bool = True,
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
    r"""Build kpoint ISDF-THC tensors.

    Given the orbitals evaluated on a (dense) real space grid, and a set of
    interpolating points (indexed  by interp_indx) determine the interpolating
    orbitals (chi), central tensor (zeta), and interpolating vectors Theta (also
    called xi).

    For the double translation case we build zeta[Q, G, G'] for all possible G
    and G' that satisfy Q - (Q-k) = G. If only_unique_G is True we only build
    the unique G's which satisfiy this expression rather than all 27^2.

    Args:
        df_inst: instance of pyscf.pbc.df.FFTDF object.
        interp_indx: array indexing interpolating points determined through
            K-Means CVT procedure. Dimension [num_interp_points]
        kpts: Array of k-points.
        orbitals: orbitals on a grid of shape [num_grid_points,
            num_orbitals], note num_orbitals = N_k * m, where m is the number of
            orbitals in the unit cell and N_k is the number of k-points.
        grid_points: Real space grid. Dimension [num_grid_points, num_dim],
            num_dim is 1, 2 or 3 for 1D, 2D, 3D.
        only_unique_G: Only build central tensor for unique Gs which satisfy
            momentum conservation condition.

    Returns:
        chi: orbitals on interpolating points
        zeta: THC central tensor. Dimension is
            [num_kpts, 27, 27, num_interp_points, num_interp_points]
            if only_unique_G is False otherwise it is of shape [num_kpts,
            num_unique[q], num_unique[q], 27, num_interp_points,
            num_interp_points].
        theta: matrix of interpolating vectors of dimension [num_grid_points,
            num_interp_points] (also called xi_mu(r)), where num_grid_points is
            the number of real space grid points and num_interp_points is the
            number of interpolating points. zeta (the
        g_mapping: g_mapping maps k-points to the appropriate delta_g index, i.e
            g_mapping[iq, ik] = i_delta_g. the index will map to the appropriate
            index in the reduced set of g vectors.
    """
    num_grid_points = len(grid_points)
    assert orbitals.shape[0] == num_grid_points
    xi, chi = solve_isdf(orbitals, interp_indx)
    momentum_map = build_momentum_transfer_mapping(df_inst.cell, kpts)
    num_kpts = len(kpts)
    num_interp_points = xi.shape[1]
    assert xi.shape == (num_grid_points, num_interp_points)
    (
        g_vectors,
        g_mapping,
        g_mapping_unique,
        delta_gs_unique,
    ) = build_g_vector_mappings_double_translation(df_inst.cell, kpts,
                                                   momentum_map)
    if only_unique_g:
        g_mapping = g_mapping_unique
        delta_gs = delta_gs_unique
    else:
        delta_gs = [g_vectors] * num_kpts
        g_mapping = g_mapping
    zeta = np.zeros((num_kpts,), dtype=object)
    for iq in range(num_kpts):
        num_g = len(delta_gs[iq])
        out_array = np.zeros(
            (num_g, num_g, num_interp_points, num_interp_points),
            dtype=np.complex128,
        )
        for ig, delta_g in enumerate(delta_gs[iq]):
            for ig_prime, delta_g_prime in enumerate(delta_gs[iq]):
                zeta_indx = build_kpoint_zeta(df_inst, kpts[iq], delta_g,
                                              delta_g_prime, grid_points, xi)
                out_array[ig, ig_prime] = zeta_indx
        zeta[iq] = out_array
    return chi, zeta, xi, g_mapping


def kpoint_isdf_single_translation(
        df_inst: df.FFTDF,
        interp_indx: npt.NDArray,
        kpts: npt.NDArray,
        orbitals: npt.NDArray,
        grid_points: npt.NDArray,
) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray, npt.NDArray]:
    r"""Build kpoint ISDF-THC tensors.

    Given the orbitals evaluated on a (dense) real space grid, and a set of
    interpolating points (indexed  by interp_indx) determine the interpolating
    orbitals (chi), central tensor (zeta), and interpolating vectors Theta (also
    called xi).

    For the double translation case we build zeta[Q, G, G'] for all possible G
    and G' that satisfy Q - (Q-k) = G. If only_unique_G is True we only build
    the unique G's which satisfiy this expression rather than all 27^2.

    Args:
        df_inst: instance of pyscf.pbc.df.FFTDF object.
        interp_indx: array indexing interpolating points determined through
            K-Means CVT procedure. Dimension [num_interp_points]
        kpts: Array of k-points.
        orbitals: orbitals on a grid of shape [num_grid_points,
            num_orbitals], note num_orbitals = N_k * m, where m is the number of
            orbitals in the unit cell and N_k is the number of k-points.
        grid_points: Real space grid. Dimension [num_grid_points, num_dim],
            num_dim is 1, 2 or 3 for 1D, 2D, 3D.

    Returns:
        chi: orbitals on interpolating points
        zeta: THC central tensor. Dimension is
            [num_kpts, 27, 27, num_interp_points, num_interp_points]
            if only_unique_G is False otherwise it is of shape [num_kpts,
            num_unique[Q], num_unique[Q], 27, num_interp_points,
            num_interp_points].
        Theta: Matrix of interpolating vectors of dimension [num_grid_points,
            num_interp_points] (also called xi_mu(r)), where num_grid_points is
            the number of real space grid points and num_interp_points is the
            number of interpolating points. Zeta (the
        G_mapping: G_mapping maps k-points to the appropriate delta_G index, i.e
            G_mapping[iq, ik] = i_delta_G. the index will map to the appropriate
            index in the reduced set of G vectors.
    """
    num_grid_points = len(grid_points)
    assert orbitals.shape[0] == num_grid_points
    xi, chi = solve_isdf(orbitals, interp_indx)
    num_kpts = len(kpts)
    num_interp_points = xi.shape[1]
    assert xi.shape == (num_grid_points, num_interp_points)
    kpts_pq = np.array([(kp, kpts[ikq])
                        for ikp, kp in enumerate(kpts)
                        for ikq in range(num_kpts)])
    kpts_pq_indx = np.array([
        (ikp, ikq) for ikp, kp in enumerate(kpts) for ikq in range(num_kpts)
    ])
    transfers = kpts_pq[:, 0] - kpts_pq[:, 1]
    unique_q, unique_indx, _ = unique(transfers)
    _, _, g_map_unique, delta_gs = build_g_vector_mappings_single_translation(
        df_inst.cell, kpts, kpts_pq_indx[unique_indx])
    num_q_vectors = len(unique_q)
    zeta = np.zeros((num_q_vectors,), dtype=object)
    for iq in range(len(unique_q)):
        num_g = len(delta_gs[iq])
        out_array = np.zeros((num_g, num_interp_points, num_interp_points),
                             dtype=np.complex128)
        for ig, delta_g in enumerate(delta_gs[iq]):
            zeta_indx = build_kpoint_zeta_single_tranlsation(
                df_inst, unique_q[iq], delta_g, grid_points, xi)
            out_array[ig] = zeta_indx
        zeta[iq] = out_array
    return chi, zeta, xi, g_map_unique


def build_isdf_orbital_inputs(mf_inst: scf.RHF) -> npt.NDArray:
    """Build orbital product inputs from mean field object

    Args:
        mf_inst: pyscf pbc mean-field object.
        mf_inst: scf.RHF:

    Returns:
        cell_periodic_mo: cell periodic part of Bloch orbital on real space grid
            of shape [num_grid_points, num_kpts*num_mo]

    """
    cell = mf_inst.cell
    kpts = mf_inst.kpts
    num_kpts = len(kpts)
    grid_points = cell.gen_uniform_grids(mf_inst.with_df.mesh)
    num_mo = mf_inst.mo_coeff[0].shape[-1]  # assuming the same for each k-point
    num_grid_points = grid_points.shape[0]
    bloch_orbitals_ao = np.array(
        numint.eval_ao_kpts(cell, grid_points, kpts=kpts))
    bloch_orbitals_mo = np.einsum("kRp,kpi->kRi",
                                  bloch_orbitals_ao,
                                  mf_inst.mo_coeff,
                                  optimize=True)
    exp_minus_ikr = np.exp(-1j * np.einsum("kx,Rx->kR", kpts, grid_points))
    cell_periodic_mo = np.einsum("kR,kRi->kRi", exp_minus_ikr,
                                 bloch_orbitals_mo)
    # go from kRi->Rki
    # AO ISDF
    cell_periodic_mo = cell_periodic_mo.transpose((1, 0, 2)).reshape(
        (num_grid_points, num_kpts * num_mo))
    return cell_periodic_mo


def density_guess(
        density: npt.NDArray,
        grid_inst: UniformGrids,
        grid_points: npt.NDArray,
        num_interp_points: int,
) -> npt.NDArray:
    """Select initial centroids based on electronic density.

    Args:
        density: Density on real space grid.
        grid_inst: pyscf UniformGrids object.
        grid_points: Real space grid points.
        num_interp_points: Number of interpolating points.

    Returns:
        grid_points: Grid points sampled using density as a weighting function.

    """
    norm_factor = np.einsum("R,R->", density, grid_inst.weights).real
    prob_dist = (density.real * grid_inst.weights) / norm_factor
    indx = np.random.choice(
        len(grid_points),
        num_interp_points,
        replace=False,
        p=prob_dist,
    )
    return grid_points[indx]


def interp_indx_from_qrcp(Z: npt.NDArray,
                          num_interp_pts: npt.NDArray,
                          return_diagonal: bool = False):
    """Find interpolating points via QRCP

    Z^T P = Q R, where R has diagonal elements that are in descending order of
    magnitude. Interpolating points are then chosen by the columns of the
    permuted Z^T.

    Args:
        Z: Orbital pair product matrix.
        num_interp_pts: integer corresponding to number of interpolating points
            to select from full real space grid.
        return_diagonal: bool:  (Default value = False)

    Returns:
        interp_indx: Index of interpolating points in full real space grid.
    """

    Q, R, P = scipy.linalg.qr(Z.T, pivoting=True)
    signs_R = np.diag(np.sign(R.diagonal()))
    # Force diagonal of R to be positive which isn't strictly enforced in QR
    # factorization.
    R = np.dot(signs_R, R)
    Q = np.dot(Q, signs_R)
    interp_indx = P[:num_interp_pts]
    if return_diagonal:
        return interp_indx, R.diagonal()
    else:
        return interp_indx


def setup_isdf(mf_inst: scf.RHF, verbose: bool = False
              ) -> Tuple[npt.NDArray, npt.NDArray, npt.NDArray]:
    """Setup common data for ISDF solution.

    Args:
        mf_inst: pyscf pbc mean-field object.
        verbose: Whether to print some information.
        mf_inst: scf.RHF:
        verbose: bool:  (Default value = False)

    Returns:
        grid_points: Real space grid points.
        cell_periodic_mo: Cell periodic part of MOs on a grid.
            [num_grid, num_kpts * num_orb]
        bloch_orbitals_mo: MOs on a grid. [num_grid, num_kpts, num_orb]

    """
    assert isinstance(mf_inst.with_df, df.FFTDF), "mf object must use FFTDF"
    cell = mf_inst.cell
    kpts = mf_inst.kpts
    grid_points = cell.gen_uniform_grids(mf_inst.with_df.mesh)
    num_grid_points = grid_points.shape[0]
    if verbose:
        print("Real space grid shape: ({}, {})".format(grid_points.shape[0],
                                                       grid_points.shape[1]))
        print("Number of grid points: {}".format(num_grid_points))
    bloch_orbitals_ao = np.array(
        numint.eval_ao_kpts(cell, grid_points, kpts=kpts))
    bloch_orbitals_mo = np.einsum("kRp,kpi->kRi",
                                  bloch_orbitals_ao,
                                  mf_inst.mo_coeff,
                                  optimize=True)
    num_mo = mf_inst.mo_coeff[0].shape[-1]  # assuming the same for each k-point
    num_kpts = len(kpts)
    # Cell periodic part
    # u = e^{-ik.r} phi(r)
    exp_minus_ikr = np.exp(-1j * np.einsum("kx,Rx->kR", kpts, grid_points))
    cell_periodic_mo = np.einsum("kR,kRi->kRi", exp_minus_ikr,
                                 bloch_orbitals_mo)
    # go from kRi->Rki
    # AO ISDF
    cell_periodic_mo = cell_periodic_mo.transpose((1, 0, 2)).reshape(
        (num_grid_points, num_kpts * num_mo))
    return grid_points, cell_periodic_mo, bloch_orbitals_mo


@dataclass
class KPointTHC:
    """Light class to hold THC tensors.

    Atributes:
        chi: THC leaf tensor. shape = [num_kpts, num_mo, num_thc]
        zeta: THC central tensor.
          shape = [num_kpts, num_G, num_G, num_thc, num_thc]
        G_mapping: For a given Q and k index this array gives the G index for
          zeta.
        xi: ISDF interpolating vectors. May be none if this class holds
            reoptimized THC decomposition.

    Examples:
        The following pseudocode shows how you can build an ERI matrix block
        given elements of this class.

        >>> kthc = solve_kmeans_kpisdf(...)
        >>> ikp, ikq, ikr, iks = kpts_indx
        >>> Gpq = kthc.G_mapping[q_indx, ikp]
        >>> Gsr = kthc.G_mapping[q_indx, iks]
        >>> eri = np.einsum(
            "pm,qm,mn,rn,sn->pqrs",
            kthc.chi[ikp].conj(),
            kthc.chi[ikq],
            kthc.zeta[q_indx][Gpq, Gsr],
            kthc.chi[ikr].conj(),
            kthc.chi[iks],
            optimize=True,
        )
    """

    chi: npt.NDArray[np.complex128]
    zeta: npt.NDArray
    g_mapping: npt.NDArray
    xi: Union[npt.NDArray[np.complex128], None]

    @property
    def num_interp_points(self) -> int:
        """Number of interpolating points (THC dimension)"""
        return self.chi.shape[-1]

    @property
    def num_thc_factors(self) -> int:
        """Number of interpolating points (THC dimension)"""
        return self.chi.shape[-1]


def solve_kmeans_kpisdf(
        mf_inst: scf.RHF,
        num_interp_points: int,
        max_kmeans_iteration: int = 500,
        single_translation: bool = False,
        use_density_guess: bool = True,
        kmeans_weighting_function: str = "density",
        verbose: bool = True,
) -> KPointTHC:
    r"""Solve for k-point THC factors using k-means CVT ISDF procedure.

    Args:
        mf_inst: pyscf pbc mean-field object.
        num_interp_points: Number of interpolating points (THC rank M).
        max_kmeans_iteration: Max iteration for k-means CVT algorithm.
        single_translation: Build THC factors assuming single translation of kp
            - kq. If true we build zeta[Q, G], else zeta[Q, G, G'].
        use_density_guess: Select initial grid points according to electron
            density? Default True.
        kmeans_weighting_function: Weighting function to use in k-means CVT.
            One of ["density", "orbital_density", "sum_squares"].
        verbose: Whether to print some information.

    Returns:
        solution: THC factors held in KPointTHC object.
    """
    if verbose:
        print(f" Number of interpolating points: {num_interp_points}")
        print(f" Max K-Means iteration: {max_kmeans_iteration}")
        print(f" Use density initial guess: {density_guess}")
        print(f" K-Means weighting function: {kmeans_weighting_function}")
        print(f" Single Translation: {single_translation}")
    # Build real space grid, and orbitals on real space grid
    grid_points, cell_periodic_mo, bloch_orbitals_mo = setup_isdf(mf_inst)
    grid_inst = UniformGrids(mf_inst.cell)
    # Find interpolating points using K-Means CVT algorithm
    kmeans = KMeansCVT(grid_points, max_iteration=max_kmeans_iteration)
    nocc = mf_inst.cell.nelec[0]  # assuming same for each k-point
    density = np.einsum(
        "kRi,kRi->R",
        bloch_orbitals_mo[:, :, :nocc].conj(),
        bloch_orbitals_mo[:, :, :nocc],
        optimize=True,
    )
    if use_density_guess:
        initial_centroids = density_guess(density, grid_inst, grid_points,
                                          num_interp_points)
    else:
        initial_centroids = None
    weighting_function = None
    # Build weighting function for CVT algorithm.
    if kmeans_weighting_function == "density":
        weighting_function = density
    elif kmeans_weighting_function == "orbital_density":
        # w(r) = sum_{ij} |phi_i(r)| |phi_j(r)|
        tmp = np.einsum(
            "kRi,pRj->Rkipj",
            abs(bloch_orbitals_mo),
            abs(bloch_orbitals_mo),
            optimize=True,
        )
        weighting_function = np.einsum("Rkipj->R", tmp)
    elif kmeans_weighting_function == "sum_squares":
        # w(r) = sum_{i} |phi_{ki}(r)|
        weighting_function = np.einsum(
            "kRi,kRi->R",
            bloch_orbitals_mo.conj(),
            bloch_orbitals_mo,
            optimize=True,
        )
        weighting_function = weighting_function
    else:
        raise ValueError(
            f"Unknown value for weighting function {kmeans_weighting_function}")
    interp_indx = kmeans.find_interpolating_points(
        num_interp_points,
        weighting_function.real,
        verbose=verbose,
        centroids=initial_centroids,
    )
    solution = solve_for_thc_factors(
        mf_inst,
        interp_indx,
        cell_periodic_mo,
        grid_points,
        single_translation=single_translation,
        verbose=verbose,
    )
    return solution


def solve_qrcp_isdf(
        mf_inst: scf.RHF,
        num_interp_points: int,
        single_translation: bool = True,
        verbose: bool = True,
) -> KPointTHC:
    r"""Solve for k-point THC factors using QRCP ISDF procedure.

    Args:
        mf_inst: pyscf pbc mean-field object.
        num_interp_points: Number of interpolating points (THC rank M).
        single_translation: Build THC factors assuming single translation of kp
        single_translation: Build THC factors assuming single translation of kp
            - kq. If true we build zeta[Q, G], else zeta[Q, G, G'].
        verbose: Whether to print some information.

    Returns:
      solution: THC factors held in KPointTHC object.
    """
    # Build real space grid, and orbitals on real space grid
    grid_points, cell_periodic_mo, _ = setup_isdf(mf_inst)
    # Find interpolating points using K-Means CVT algorithm
    # Z_{R, (ki)(k'j)} = u_{ki}(r)* u_{k'j}(r)
    num_grid_points = len(grid_points)
    num_orbs = cell_periodic_mo.shape[1]
    pair_products = np.einsum("Ri,Rj->Rij",
                              cell_periodic_mo.conj(),
                              cell_periodic_mo,
                              optimize=True).reshape(
                                  (num_grid_points, num_orbs**2))
    interp_indx = interp_indx_from_qrcp(pair_products, num_interp_points)
    # Solve for THC factors.
    solution = solve_for_thc_factors(
        mf_inst,
        interp_indx,
        cell_periodic_mo,
        grid_points,
        single_translation=single_translation,
        verbose=verbose,
    )
    return solution


def solve_for_thc_factors(
        mf_inst,
        interp_points_index,
        cell_periodic_mo,
        grid_points,
        single_translation=True,
        verbose=True,
) -> KPointTHC:
    r"""Solve for k-point THC factors using interpolating points as input.

    Args:
        mf_inst: pyscf pbc mean-field object.
        interp_points_index: Indices of interpolating points found from k-means
            CVT or QRCP.
        cell_periodic_mo: cell periodic part of Bloch orbital on real space grid
            shape: [num_grid_points, num_kpts*num_mo]
        kmeans_weighting_function: Weighting function to use in k-means CVT.
            One of ["density", "orbital_density", "sum_squares"].
        grid_points: Real space grid. Dimension [num_grid_points, num_dim],
                num_dim is 1, 2 or 3 for 1D, 2D, 3D.
        single_translation: Build THC factors assuming single translation of kp
            - kq. If true we build zeta[Q, G], else zeta[Q, G, G'].
            (Default value = True)
        verbose: Whether to print some information. (Default value = True)

    Returns:
        solution: THC factors held in KPointTHC object.
    """
    assert isinstance(mf_inst.with_df, df.FFTDF), "mf object must use FFTDF"
    if single_translation:
        chi, zeta, xi, g_mapping = kpoint_isdf_single_translation(
            mf_inst.with_df,
            interp_points_index,
            mf_inst.kpts,
            cell_periodic_mo,
            grid_points,
        )
    else:
        chi, zeta, xi, g_mapping = kpoint_isdf_double_translation(
            mf_inst.with_df,
            interp_points_index,
            mf_inst.kpts,
            cell_periodic_mo,
            grid_points,
        )
    num_interp_points = len(interp_points_index)
    num_kpts = len(mf_inst.kpts)
    num_mo = mf_inst.mo_coeff[0].shape[-1]  # assuming the same for each k-point
    assert chi.shape == (num_interp_points, num_kpts * num_mo)
    # go from Rki -> kiR
    chi = chi.reshape((num_interp_points, num_kpts, num_mo))
    chi = chi.transpose((1, 2, 0))

    solution = KPointTHC(chi=chi, zeta=zeta, xi=xi, g_mapping=g_mapping)

    return solution



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from typing import Union
import numpy as np
import numpy.typing as npt

from pyscf.pbc import scf
from pyscf.pbc.lib.kpts_helper import unique, member

from openfermion.resource_estimates.pbc.hamiltonian import (
    build_momentum_transfer_mapping,)

from openfermion.resource_estimates.pbc.thc.factorizations.isdf import (
    build_g_vector_mappings_double_translation,
    build_g_vector_mappings_single_translation,
    build_eri_isdf_double_translation,
    build_eri_isdf_single_translation,
)


class KPTHCDoubleTranslation:

    def __init__(
            self,
            chi: npt.NDArray,
            zeta: npt.NDArray,
            kmf: scf.HF,
            chol: Union[npt.NDArray, None] = None,
    ):
        """Class for constructing THC factorized ERIs.

        Assumes 2 G vectors to index THC factors (i.e. `double translation.')

        Arguments:
            chi: array of interpolating orbitals of shape
                [num_kpts, num_mo, num_interp_points]
            zeta: central tensor of dimension
                [num_kpts, num_G, num_G, num_interp_points, num_interp_points].
            kmf: pyscf k-object.  Currently only used to obtain the number of
                k-points.  must have an attribute kpts which len(self.kmf.kpts)
                returns number of kpts.
            cholesky_factor: Cholesky object for computing exact integrals
        """
        self.chi = chi
        self.zeta = zeta
        self.kmf = kmf
        self.nk = len(self.kmf.kpts)
        self.kpts = self.kmf.kpts
        self.k_transfer_map = build_momentum_transfer_mapping(
            self.kmf.cell, self.kmf.kpts)
        self.reverse_k_transfer_map = np.zeros_like(
            self.k_transfer_map)  # [kidx, kmq_idx] = qidx
        for kidx in range(self.nk):
            for qidx in range(self.nk):
                kmq_idx = self.k_transfer_map[qidx, kidx]
                self.reverse_k_transfer_map[kidx, kmq_idx] = qidx
        # Two-translation ISDF zeta[iq, dG, dG']
        _, _, g_map_unique, _ = build_g_vector_mappings_double_translation(
            self.kmf.cell, self.kmf.kpts, self.k_transfer_map)
        self.g_mapping = g_map_unique
        self.chol = chol

    def get_eri(self, ikpts: list) -> npt.NDArray:
        r"""Construct ERIs given kpt indices.

        .. math::

            (pkp qkq| rkr sks) = \\sum_{mu nu} zeta[iq, dG, dG', mu, nu]
            chi[kp,p,mu]* chi[kq,q,mu] chi[kp,p,nu]* chi[ks,s,nu]

        Arguments:
            ikpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, _, _ = ikpts
        q_indx = self.reverse_k_transfer_map[ikp, ikq]
        return build_eri_isdf_double_translation(self.chi, self.zeta, q_indx,
                                                 ikpts, self.g_mapping)

    def get_eri_exact(self, kpts: list) -> npt.NDArray:
        """Construct (pkp qkq| rkr sks) exactly from cholesky factors.

        Arguments:
            kpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = kpts
        if self.chol is not None:
            return np.einsum(
                "npq,nsr->pqrs",
                self.chol[ikp, ikq],
                self.chol[iks, ikr].conj(),
                optimize=True,
            )
        else:
            eri_kpt = self.kmf.with_df.ao2mo(
                [self.kmf.mo_coeff[i] for i in (ikp, ikq, ikr, iks)],
                [self.kmf.kpts[i] for i in (ikp, ikq, ikr, iks)],
                compact=False,
            )
            shape_pqrs = [
                self.kmf.mo_coeff[i].shape[-1] for i in (ikp, ikq, ikr, iks)
            ]
            eri_kpt = eri_kpt.reshape(shape_pqrs)
        return eri_kpt


class KPTHCSingleTranslation(KPTHCDoubleTranslation):

    def __init__(
            self,
            chi: npt.NDArray,
            zeta: npt.NDArray,
            kmf: scf.HF,
    ):
        """Class for constructing THC factorized ERIs.

        Assumes one delta G (i.e. a single translation vector.)

        Arguments:
            chi: array of interpolating orbitals of shape
                [num_kpts, num_mo, num_interp_points]
            zeta: central tensor of dimension
                [num_kpts, num_G, num_G, num_interp_points, num_interp_points].
            kmf: pyscf k-object.  Currently only used to obtain the number of
                k-points. must have an attribute kpts which len(self.kmf.kpts)
                returns number of kpts.
            cholesky_factor: Cholesky object for computing exact integrals
        """
        super().__init__(chi, zeta, kmf)
        # one-translation ISDF zeta[iq, dG]
        num_kpts = len(self.kmf.kpts)
        kpts = self.kmf.kpts
        kpts_pq = np.array([(kp, kpts[ikq])
                            for ikp, kp in enumerate(kpts)
                            for ikq in range(num_kpts)])
        kpts_pq_indx = np.array([
            (ikp, ikq) for ikp, kp in enumerate(kpts) for ikq in range(num_kpts)
        ])
        transfers = kpts_pq[:, 0] - kpts_pq[:, 1]
        _, unique_indx, _ = unique(transfers)
        (
            _,
            _,
            g_map_unique,
            _,
        ) = build_g_vector_mappings_single_translation(
            kmf.cell, kpts, kpts_pq_indx[unique_indx])
        self.g_mapping = g_map_unique
        self.momentum_transfers = transfers[unique_indx]

    def get_eri(self, ikpts):
        """Construct ERIs from kpt indices

        Evaluated via

        .. math::

            (pkp qkq| rkr sks) = \\sum_{mu nu} zeta[iq, dG, mu, nu]
                chi[kp,p,mu]* chi[kq,q,mu] chi[kp,p,nu]* chi[ks,s,nu]

        Arguments:
            kpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, _, _ = ikpts
        mom_transfer = self.kpts[ikp] - self.kpts[ikq]
        q_indx = member(mom_transfer, self.momentum_transfers)[0]
        return build_eri_isdf_single_translation(self.chi, self.zeta, q_indx,
                                                 ikpts, self.g_mapping)

    def get_eri_exact(self, kpts):
        """Construct (pkp qkq| rkr sks) exactly from cholesky factors.

        Arguments:
            kpts: list of four integers representing the index of the kpoint in
                self.kmf.kpts

        Returns:
            eris: ([pkp][qkq]|[rkr][sks])
        """
        ikp, ikq, ikr, iks = kpts
        if self.chol is not None:
            return np.einsum(
                "npq,nsr->pqrs",
                self.chol[ikp, ikq],
                self.chol[iks, ikr].conj(),
                optimize=True,
            )
        else:
            eri_kpt = self.kmf.with_df.ao2mo(
                [self.kmf.mo_coeff[i] for i in (ikp, ikq, ikr, iks)],
                [self.kmf.kpts[i] for i in (ikp, ikq, ikr, iks)],
                compact=False,
            )
            shape_pqrs = [
                self.kmf.mo_coeff[i].shape[-1] for i in (ikp, ikq, ikr, iks)
            ]
            eri_kpt = eri_kpt.reshape(shape_pqrs)
        return eri_kpt



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import mp, cc
    from openfermion.resource_estimates.pbc.hamiltonian.cc_extensions import (
        build_approximate_eris,)

    from openfermion.resource_estimates.pbc.sf.sf_integrals import (
        SingleFactorization,)
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.testing import (
        make_diamond_113_szv,)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_sf_helper_trunc():
    mf = make_diamond_113_szv()
    exact_cc = cc.KRCCSD(mf)
    eris = exact_cc.ao2mo()
    exact_emp2, _, _ = exact_cc.init_amps(eris)

    mymp = mp.KMP2(mf)

    Luv = cholesky_from_df_ints(mymp)  # [kpt, kpt, naux, nao, nao]
    naux = Luv[0, 0].shape[0]

    print(" naux  error (Eh)")
    approx_cc = cc.KRCCSD(mf)
    approx_cc.verbose = 0
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=10)

    eris = build_approximate_eris(approx_cc, helper)
    emp2, _, _ = approx_cc.init_amps(eris)
    assert not np.isclose(emp2, exact_emp2)

    out_eris = build_approximate_eris(approx_cc, helper)
    emp2_2, _, _ = approx_cc.init_amps(out_eris)
    assert not np.isclose(emp2, exact_emp2)
    assert np.isclose(emp2, emp2_2)
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=5)
    out_eris = build_approximate_eris(approx_cc, helper)
    emp2_2, _, _ = approx_cc.init_amps(out_eris)
    assert not np.isclose(emp2, exact_emp2)
    assert not np.isclose(emp2, emp2_2)
    out_eris = build_approximate_eris(approx_cc, helper, eris=eris)
    emp2_3, _, _ = approx_cc.init_amps(out_eris)
    assert not np.isclose(emp2, exact_emp2)
    assert np.isclose(emp2_2, emp2_3)

    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=naux)
    out_eris = build_approximate_eris(approx_cc, helper)
    emp2, _, _ = approx_cc.init_amps(out_eris)
    assert np.isclose(emp2, exact_emp2)



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os

import pytest
import numpy as np

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from pyscf.pbc import gto, scf, mp
    from pyscf.pbc.cc import KRCCSD
    from pyscf.lib import chkfile
    from openfermion.resource_estimates.pbc.hamiltonian import (
        cholesky_from_df_ints,)
    from openfermion.resource_estimates.pbc.hamiltonian.cc_extensions import (
        build_approximate_eris_rohf,
        build_approximate_eris,
    )
    from openfermion.resource_estimates.pbc.sf.sf_integrals import (
        SingleFactorization,)

_TEST_CHK = os.path.join(os.path.dirname(__file__), "../test_data/scf.chk")


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_cc_extension_rohf():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.build()

    kmesh = [1, 1, 3]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KROHF(cell, kpts).rs_density_fit()
    mf.with_df.mesh = cell.mesh
    mf.kernel()

    # Only ROHF integrals are supported for resource estimates but only UCCSD
    # available, so convert MOs to UHF format first.
    u_from_ro = scf.addons.convert_to_uhf(mf)
    mymp = mp.KMP2(mf)

    Luv = cholesky_from_df_ints(mymp)  # [kpt, kpt, naux, nao, nao]
    from pyscf.pbc.cc.kccsd_uhf import _make_df_eris
    from pyscf.pbc.cc import KUCCSD

    cc_inst = KUCCSD(u_from_ro)
    eris = _make_df_eris(cc_inst)
    emp2_ref, _, _ = cc_inst.init_amps(eris)
    ref_eris = _make_df_eris(cc_inst)

    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf)
    test_eris = build_approximate_eris_rohf(cc_inst, helper)
    eri_blocks = [
        "OOOO",
        "OOOV",
        "OOVV",
        "OVOV",
        "VOOV",
        "VOVV",
        "oooo",
        "ooov",
        "oovv",
        "ovov",
        "voov",
        "vovv",
        "OOov",
        "OOvv",
        "OVov",
        "VOov",
        "VOvv",
    ]
    for block in eri_blocks:
        assert np.allclose(test_eris.__dict__[block][:],
                           ref_eris.__dict__[block][:])
    # Test MP2 energy is the correct
    emp2_approx, _, _ = cc_inst.init_amps(test_eris)
    assert abs(emp2_approx - emp2_ref) < 1e-12

    # Test MP2 energy is the different when truncated
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=10)
    test_eris_approx = build_approximate_eris_rohf(cc_inst, helper)
    for block in eri_blocks:
        assert not np.allclose(test_eris_approx.__dict__[block][:],
                               ref_eris.__dict__[block][:])
    emp2_approx, _, _ = cc_inst.init_amps(test_eris_approx)
    # MP2 energy should be pretty bad
    assert abs(emp2_approx - emp2_ref) > 1e-12
    # Test CCSD
    cc_exact = KUCCSD(u_from_ro)
    ecc_exact, _, _ = cc_exact.kernel()
    cc_approx = KUCCSD(u_from_ro)
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf)
    test_eris = build_approximate_eris_rohf(cc_approx, helper)
    cc_approx.ao2mo = lambda mo_coeff=None: test_eris
    emp2_approx, _, _ = cc_approx.init_amps(test_eris)
    ecc_approx, _, _ = cc_approx.kernel()
    assert abs(ecc_exact - ecc_approx) < 1e-12
    # Should be different
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=10)
    test_eris = build_approximate_eris_rohf(cc_approx, helper)
    for block in eri_blocks:
        assert not np.allclose(test_eris.__dict__[block][:],
                               ref_eris.__dict__[block][:])
    # Want to avoid ccsd helper using density-fitted integrals which will be
    # "exact"
    assert test_eris.Lpv is None
    assert test_eris.LPV is None
    cc_approx = KUCCSD(u_from_ro)
    # overwrite ao2mo object required as function does not check if eris exists.
    cc_approx.ao2mo = lambda mo_coeff=None: test_eris
    ecc_approx, _, _ = cc_approx.kernel()
    assert abs(ecc_exact - ecc_approx) > 1e-12


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
@pytest.mark.slow
def test_cc_helper_rhf():
    cell = gto.Cell()
    cell.atom = """
    C 0.000000000000   0.000000000000   0.000000000000
    C 1.685068664391   1.685068664391   1.685068664391
    """
    cell.basis = "gth-szv"
    cell.pseudo = "gth-hf-rev"
    cell.a = """
    0.000000000, 3.370137329, 3.370137329
    3.370137329, 0.000000000, 3.370137329
    3.370137329, 3.370137329, 0.000000000"""
    cell.unit = "B"
    cell.verbose = 0
    cell.build()

    kmesh = [1, 1, 3]
    kpts = cell.make_kpts(kmesh)
    mf = scf.KRHF(cell, kpts).rs_density_fit()
    scf_dict = chkfile.load(_TEST_CHK, "scf")
    mf.__dict__.update(scf_dict)
    mf.with_df._cderi = _TEST_CHK
    dm0 = mf.make_rdm1()
    mf.with_df.mesh = cell.mesh
    mf.kernel(dm0=dm0)

    # Only ROHF integrals are supported for resource estimates but only UCCSD
    # available, so convert MOs to UHF format first.
    mymp = mp.KMP2(mf)

    Luv = cholesky_from_df_ints(mymp)  # [kpt, kpt, naux, nao, nao]

    cc_inst = KRCCSD(mf)
    ref_eris = cc_inst.ao2mo()
    emp2_ref, _, _ = cc_inst.init_amps(ref_eris)

    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf)
    test_eris = build_approximate_eris(cc_inst, helper)
    eri_blocks = [
        "oooo",
        "ooov",
        "oovv",
        "ovov",
        "voov",
        "vovv",
    ]
    for block in eri_blocks:
        assert np.allclose(test_eris.__dict__[block][:],
                           ref_eris.__dict__[block][:])
    # Test MP2 energy is the correct
    emp2_approx, _, _ = cc_inst.init_amps(test_eris)
    assert abs(emp2_approx - emp2_ref) < 1e-12

    # Test MP2 energy is the different when truncated
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=10)
    test_eris_approx = build_approximate_eris(cc_inst, helper)
    for block in eri_blocks:
        assert not np.allclose(test_eris_approx.__dict__[block][:],
                               ref_eris.__dict__[block][:])
    emp2_approx, _, _ = cc_inst.init_amps(test_eris_approx)
    # MP2 energy should be pretty bad
    assert abs(emp2_approx - emp2_ref) > 1e-12
    # Test CCSD
    cc_exact = KRCCSD(mf)
    ecc_exact, _, _ = cc_exact.kernel()
    cc_approx = KRCCSD(mf)
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf)
    test_eris = build_approximate_eris(cc_approx, helper)
    cc_approx.ao2mo = lambda mo_coeff=None: test_eris
    emp2_approx, _, _ = cc_approx.init_amps(test_eris)
    ecc_approx, _, _ = cc_approx.kernel()
    assert abs(ecc_exact - ecc_approx) < 1e-12
    # Should be different
    helper = SingleFactorization(cholesky_factor=Luv, kmf=mf, naux=10)
    test_eris = build_approximate_eris(cc_approx, helper)
    for block in eri_blocks:
        assert not np.allclose(test_eris.__dict__[block][:],
                               ref_eris.__dict__[block][:])
    # Want to avoid ccsd helper using density-fitted integrals which will be
    # "exact"
    cc_approx = KRCCSD(mf)
    # overwrite ao2mo object required as function does not check if eris exists.
    cc_approx.ao2mo = lambda mo_coeff=None: test_eris
    ecc_approx, _, _ = cc_approx.kernel()
    assert abs(ecc_exact - ecc_approx) > 1e-12



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
"""Module for testing how to map k-points between each other.

Reciprocal lattice vectors (G_{pqr}) are defined as linear combinations of
underlying primitive vectors {b_i} via

.. math::

    G_{pqr} = p b_1 + q b_2 + r b_3 (*)

For quantum algorithms it is convenient to work with the underlying integer
representation of reciprocal lattice vectors (pqr) (G-vectors or gvecs) rather
than vectors given by the literal sum in (*).

This module tests out various mappings between k-points and reciprocal lattice
vectors in terms of these integer values.
"""
import itertools
import numpy as np
from pyscf.lib.numpy_helper import cartesian_prod


def get_miller_indices(kmesh):
    """Calculate the miller indices from kmesh.

    Assumes a non gamma centered non-1stBZ Monkorhst-Pack mesh

    Args:
        kmesh: A length 3 1-D iteratble with the number of k-points in the x,y,z
            direction [Nk_x, NK_y, NK_z] where NK_x/y/z are positive integers

    Returns:
        np.array 2D that is prod([Nk_x, NK_y, NK_z])
    """
    assert len(kmesh) == 3, "kmesh must be of length 3."
    nx, ny, nz = kmesh
    if nx < 1:
        raise TypeError("Bad miller index dimension in x")
    if ny < 1:
        raise TypeError("Bad miller index dimension in y")
    if nz < 1:
        raise TypeError("Bad miller index dimension in z")

    ks_int_each_axis = []
    for n in kmesh:
        ks_int_each_axis.append(np.arange(n, dtype=float))
    int_scaled_kpts = cartesian_prod(ks_int_each_axis)
    return int_scaled_kpts


def get_delta_kp_kq_q(int_scaled_kpts):
    """Generate kp - kq - Q = S for kp, kq, and Q.

    The difference of the three integers is stored as a four tensor D_{kp, kq,
    Q} = S. where the dimension of D is (nkpts, nkpts, nkpts, 3).  The last
    dimension stores the x,y,z components of S.

    Args:
        int_scaled_kpts: array of kpts represented as miller indices
            [[nkx, nky, nkz], ...]

    Returns:
        np.ndarray mapping D_{kp, kq, Q}.
    """
    delta_k1_k2_q_int = (int_scaled_kpts[:, None, None, :] -
                         int_scaled_kpts[None, :, None, :] -
                         int_scaled_kpts[None, None, :, :])
    return delta_k1_k2_q_int


def build_transfer_map(kmesh, scaled_kpts):
    """Define mapping momentum_transfer_map[Q][k1] = k2.

    Where k1 - k2 + G = Q.  Here k1, k2, Q are all tuples of integers [[0,
    Nkx-1], [0, Nky-1], [0, Nkz-1]] and G is [{0, Nkx}, {0, Nky}, {0, Nkz}].

    This is computed from `get_delta_kp_kq_q` which computes k1 - k2 - Q = S.
    Thus k1 - k2 = Q + S which shows that S is [{0, Nkx}, {0, Nky}, {0, Nkz}].

    Args:
        kmesh: [nkx, nky, nkz] the number of kpoints in each direction
        scaled_kpts: miller index representation
            [[0, nkx-1], [0, nky-1], [0, nkz-1]] of all the kpoints

    Returns:
        transfer map satisfying k1 - k2 + G = Q in matrix form map[Q, k1] = k2
    """
    nkpts = len(scaled_kpts)
    delta_k1_k2_q_int = get_delta_kp_kq_q(scaled_kpts)
    transfer_map = np.zeros((nkpts, nkpts), dtype=np.int32)
    for kpidx, kqidx, qidx in itertools.product(range(nkpts), repeat=3):
        # explicitly build my transfer matrix
        if np.allclose(
            [
                np.rint(delta_k1_k2_q_int[kpidx, kqidx, qidx][0]) % kmesh[0],
                np.rint(delta_k1_k2_q_int[kpidx, kqidx, qidx][1]) % kmesh[1],
                np.rint(delta_k1_k2_q_int[kpidx, kqidx, qidx][2]) % kmesh[2],
            ],
                0,
        ):
            transfer_map[qidx, kpidx] = kqidx
    return transfer_map


def build_conjugate_map(kmesh, scaled_kpts):
    """Build mapping such that map[k1] = -k1

    Args:
        kmesh: kpoint mesh
        scaled_kpts: integer k-points

    Returns:
        kconj_map: conjugate k-point mapping
    """
    nkpts = len(scaled_kpts)
    kpoint_dict = dict(
        zip(
            [tuple(map(int, scaled_kpts[x])) for x in range(nkpts)],
            range(nkpts),
        ))
    kconj_map = np.zeros((nkpts), dtype=int)
    for kidx in range(nkpts):
        negative_k_scaled = -scaled_kpts[kidx]
        fb_negative_k_scaled = tuple((
            int(negative_k_scaled[0]) % kmesh[0],
            int(negative_k_scaled[1]) % kmesh[1],
            int(negative_k_scaled[2]) % kmesh[2],
        ))
        kconj_map[kidx] = kpoint_dict[fb_negative_k_scaled]
    return kconj_map


def build_G_vectors(kmesh):
    """Build all 8 Gvectors.

    Args:
        kmesh: kpoint mesh

    Returns:
        g_dict: a dictionary mapping miller index to appropriate
            G_vector index.  The actual cell Gvector can be recovered with
            np.einsum("x,wx->w", (n1, n2, n3), cell.reciprocal_vectors()
    """
    g_dict = {}
    indx = 0
    for n1, n2, n3 in itertools.product([0, -1], repeat=3):
        g_dict[(n1 * kmesh[0], n2 * kmesh[1], n3 * kmesh[2])] = indx
        indx += 1
    return g_dict


def build_gpq_mapping(kmesh, int_scaled_kpts):
    """Build map for kp - kq = Q + G

    Here G is [{0, -Nkx}, {0, -Nky}, {0, -Nkz}]. G will be 0 or Nkz because kp -
    kq takes on values between [-Nka + 1, Nka - 1] in each component.

    Args:
        kmesh: number of k-points along each direction [Nkx, Nky, Nkz].
        int_scaled_kpts: scaled_kpts. Each kpoint is a tuple of 3 integers
            where each integer is between [0, Nka-1].

    Returns:
        gpq_mapping: array mapping where first two indices are the index of kp
            and kq and the last dimension holds the gval that is [{0, Nkx}, {0,
            Nky}, {0, Nkz}].

    """
    momentum_map = build_transfer_map(kmesh, int_scaled_kpts)
    nkpts = len(int_scaled_kpts)
    gpq_mapping = np.zeros((nkpts, nkpts, 3), dtype=np.int32)
    for iq in range(nkpts):
        for ikp in range(nkpts):
            ikq = momentum_map[iq, ikp]
            q_minus_g = int_scaled_kpts[ikp] - int_scaled_kpts[ikq]
            g_val = (
                0 if q_minus_g[0] >= 0 else -kmesh[0],
                0 if q_minus_g[1] >= 0 else -kmesh[1],
                0 if q_minus_g[2] >= 0 else -kmesh[2],
            )
            gpq_mapping[ikp, ikq, :] = np.array(g_val)

    return gpq_mapping



# coverage: ignore
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import numpy as np
import pytest

from openfermion.resource_estimates import HAVE_DEPS_FOR_RESOURCE_ESTIMATES

if HAVE_DEPS_FOR_RESOURCE_ESTIMATES:
    from openfermion.resource_estimates.pbc.sf.compute_sf_resources import (
        _compute_cost,
        compute_cost,
    )


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_estimate():
    n = 152
    lam = 3071.8
    L = 275
    dE = 0.001
    chi = 10

    res = _compute_cost(n, lam, L, dE, chi, 20_000, 3, 3, 3)
    # 1663687, 8027577592851, 438447}
    assert np.isclose(res[0], 1663687)
    assert np.isclose(res[1], 8027577592851)
    assert np.isclose(res[2], 438447)
    res = _compute_cost(n, lam, L, dE, chi, res[0], 3, 3, 3)
    assert np.isclose(res[0], 1663707)
    assert np.isclose(res[1], 8027674096311)
    assert np.isclose(res[2], 438452)

    res = _compute_cost(n, lam, L, dE, chi, 20_000, 3, 5, 1)
    # 907828, 4380427154244, 219526
    assert np.isclose(res[0], 907828)
    assert np.isclose(res[1], 4380427154244)
    assert np.isclose(res[2], 219526)
    res = _compute_cost(n, lam, L, dE, chi, res[0], 3, 5, 1)
    assert np.isclose(res[0], 907828)
    assert np.isclose(res[1], 4380427154244)
    assert np.isclose(res[2], 219526)


@pytest.mark.skipif(not HAVE_DEPS_FOR_RESOURCE_ESTIMATES,
                    reason='pyscf and/or jax not installed.')
def test_estimate_helper():
    n = 152
    lam = 3071.8
    L = 275
    dE = 0.001
    chi = 10

    res = compute_cost(
        num_spin_orbs=n,
        lambda_tot=lam,
        num_aux=L,
        kmesh=[3, 3, 3],
        dE_for_qpe=dE,
        chi=chi,
    )
    assert np.isclose(res.toffolis_per_step, 1663707)
    assert np.isclose(res.total_toffolis, 8027674096311)
    assert np.isclose(res.logical_qubits, 438452)

    res = compute_cost(
        num_spin_orbs=n,
        lambda_tot=lam,
        num_aux=L,
        kmesh=[3, 5, 1],
        dE_for_qpe=dE,
        chi=chi,
    )
    # 1663687, 8027577592851, 438447}
    assert np.isclose(res.toffolis_per_step, 907828)
    assert np.isclose(res.total_toffolis, 4380427154244)
    assert np.isclose(res.logical_qubits, 219526)
